# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import codecs
from datetime import datetime
from fractions import Fraction
import locale
import logging
import math
import os
import re

try:
    import pgi
    pgi.install_as_gi()
    using_pgi = True
except ImportError:
    using_pgi = False
import gi
for gexiv2_vsn in ('0.10', '0.4'):
    try:
        gi.require_version('GExiv2', gexiv2_vsn)
        break
    except ValueError:
        pass
from gi.repository import GLib, GObject, GExiv2
import six

from photini import __version__
from photini.pyqt import QtCore, QtGui

logger = logging.getLogger(__name__)

gexiv2_version = '{} {}, GExiv2 {}.{}.{}, GObject {}'.format(
    ('PyGI', 'pgi')[using_pgi], gi.__version__, GExiv2.MAJOR_VERSION,
    GExiv2.MINOR_VERSION, GExiv2.MICRO_VERSION, GObject._version)

# pydoc gi.repository.GExiv2.Metadata is useful to see methods available

GExiv2.initialize()

# we can't reroute GExiv2 messages to Python logging, so mute them...
GExiv2.log_set_level(GExiv2.LogLevel.MUTE)

# ...unless in test mode
def debug_metadata():
    GExiv2.log_set_level(GExiv2.LogLevel.INFO)

# recent versions of Exiv2 have these namespaces defined, but older versions
# may not recognise them
for prefix, name in (
        ('video',   'http://www.video/'),
        ('xmpGImg', 'http://ns.adobe.com/xap/1.0/g/img/')):
    GExiv2.Metadata.register_xmp_namespace(name, prefix)

def safe_fraction(value):
    # Avoid ZeroDivisionError when '0/0' used for zero values in Exif
    if isinstance(value, six.string_types):
        numerator, sep, denominator = value.partition('/')
        if denominator and int(denominator) == 0:
            return Fraction(0.0)
    return Fraction(value).limit_denominator(1000000)

def decode_UCS2(value):
    value = bytearray(map(int, value.split()))
    return value.decode('utf_16').strip('\x00')

class MD_Value(object):
    # mixin for "metadata objects" - Python types with additional functionality
    def __nonzero__(self):
        # reinterpret as "has a value"
        return True

    def log_merged(self, info, tag, value):
        logger.info('%s: merged %s', info, tag)

    def log_replaced(self, info, tag, value):
        logger.warning(
            '%s: "%s" replaced by %s "%s"', info, str(self), tag, str(value))

    def log_ignored(self, info, tag, value):
        logger.warning('%s: ignored %s "%s"', info, tag, str(value))


class MD_Dict(dict, MD_Value):
    def __init__(self, value):
        # can initialise from a string containing comma separated values
        if isinstance(value, six.string_types):
            value = value.split(',')
        # or a list of values
        if isinstance(value, (tuple, list)):
            value = zip(self._keys, value)
        super(MD_Dict, self).__init__(value)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(
            "{} has no attribute {}".format(self.__class__, name))

    def __setattr__(self, name, value):
        if name in self:
            self[name] = value
            return
        super(MD_Dict, self).__setattr__(name, value)

    def __nonzero__(self):
        return any(self.values())

    def merge(self, info, tag, other):
        if other == self:
            return self
        ignored = False
        for key in self:
            if not other[key]:
                continue
            if not self[key]:
                self[key] = other[key]
            elif other[key] != self[key]:
                ignored = True
        if ignored:
            self.log_ignored(info, tag, other)
        else:
            self.log_merged(info, tag, other)
        return self


class FocalLength(MD_Dict):
    # store actual focal length and 35mm film equivalent
    _keys = ('fl', 'fl_35')

    def __init__(self, value):
        super(FocalLength, self).__init__(value)
        if self.fl:
            self.fl = safe_fraction(self.fl)
        if self.fl_35:
            self.fl_35 = int(self.fl_35)

    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not any(file_value):
            return None
        return cls(file_value + [None])

    def write(self, handler, tag):
        file_value = [None, None]
        if self.fl:
            file_value[0] = '{:d}/{:d}'.format(
                self.fl.numerator, self.fl.denominator)
        if self.fl_35:
            file_value[1] = '{:d}'.format(self.fl_35)
        handler.set_string(tag, file_value)

    def to_35(self, value):
        if value and self.fl and self.fl_35:
            return int((float(value) * float(self.fl_35) / self.fl) + 0.5)
        return self.fl_35

    def from_35(self, value):
        if value and self.fl and self.fl_35:
            return round(float(value) * self.fl / float(self.fl_35), 2)
        return self.fl


class LatLon(MD_Dict):
    # simple class to store latitude and longitude
    _keys = ('lat', 'lon')

    def __init__(self, value):
        super(LatLon, self).__init__(value)
        self.lat = round(float(self.lat), 6)
        self.lon = round(float(self.lon), 6)

    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if tag == 'Xmp.video.GPSCoordinates':
            if file_value:
                match = re.match(r'([-+]\d+\.\d+)([-+]\d+\.\d+)', file_value)
                if match:
                    return cls(match.group(1, 2))
            return None
        if not all(file_value):
            return None
        if handler.is_exif_tag(tag[0]):
            return cls((cls.from_exif_part(file_value[0], file_value[1]),
                        cls.from_exif_part(file_value[2], file_value[3])))
        else:
            return cls((cls.from_xmp_part(file_value[0]),
                        cls.from_xmp_part(file_value[1])))

    def write(self, handler, tag):
        if handler.is_exif_tag(tag[0]):
            lat_string, negative = self.to_exif_part(self.lat)
            lat_ref = 'NS'[negative]
            lon_string, negative = self.to_exif_part(self.lon)
            lon_ref = 'EW'[negative]
            handler.set_string(tag, (lat_string, lat_ref, lon_string, lon_ref))
        else:
            lat_string, negative = self.to_xmp_part(self.lat)
            lat_string += 'NS'[negative]
            lon_string, negative = self.to_xmp_part(self.lon)
            lon_string += 'EW'[negative]
            handler.set_string(tag, (lat_string, lon_string))

    @staticmethod
    def from_exif_part(value, ref):
        parts = [float(Fraction(x)) for x in value.split()] + [0.0, 0.0]
        result = parts[0] + (parts[1] / 60.0) + (parts[2] / 3600.0)
        if ref in ('S', 'W'):
            result = -result
        return result

    @staticmethod
    def to_exif_part(value):
        negative = value < 0.0
        if negative:
            value = -value
        degrees = int(value)
        value = (value - degrees) * 60.0
        minutes = int(value)
        seconds = (value - minutes) * 60.0
        seconds = safe_fraction(seconds)
        return '{:d}/1 {:d}/1 {:d}/{:d}'.format(
            degrees, minutes, seconds.numerator, seconds.denominator), negative

    @staticmethod
    def from_xmp_part(value):
        ref = value[-1]
        if ref in ('N', 'S', 'E', 'W'):
            value = value[:-1]
        if ',' in value:
            degrees, minutes = value.split(',')
            value = float(degrees) + (float(minutes) / 60.0)
        else:
            value = float(value)
        if ref in ('S', 'W'):
            value = -value
        return value

    @staticmethod
    def to_xmp_part(value):
        negative = value < 0.0
        if negative:
            value = -value
        degrees = int(value)
        minutes = (value - degrees) * 60.0
        return '{:d},{:.6f}'.format(degrees, minutes), negative

    def __str__(self):
        return '{:.6f}, {:.6f}'.format(self.lat, self.lon)

    def merge(self, info, tag, other):
        if max(abs(other.lat - self.lat), abs(other.lon - self.lon)) > 0.0000015:
            self.log_ignored(info, tag, other)
        return self


class Location(MD_Dict):
    # stores IPTC defined location heirarchy
    _keys = ('sublocation', 'city', 'province_state',
             'country_name', 'country_code', 'world_region')

    def __init__(self, value):
        super(Location, self).__init__(value)
        if self.country_code:
            self.country_code = self.country_code.upper()

    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not any(file_value):
            return None
        return cls(file_value + [None])

    def write(self, handler, tag):
        handler.set_string(tag, [self[x] for x in self._keys])

    def merge(self, info, tag, other):
        merged = False
        for key in self:
            if not other[key]:
                continue
            if not self[key]:
                self[key] = other[key]
                merged = True
            elif other[key] not in self[key]:
                self[key] += ' // ' + other[key]
                merged = True
        if merged:
            self.log_merged(info, tag, other)
        return self


class LensSpec(MD_Dict):
    # simple class to store lens "specificaton"
    _keys = ('min_fl', 'max_fl', 'min_fl_fn', 'max_fl_fn')

    def __init__(self, value):
        super(LensSpec, self).__init__(value)
        self.min_fl = safe_fraction(self.min_fl)
        self.max_fl = safe_fraction(self.max_fl)
        self.min_fl_fn = safe_fraction(self.min_fl_fn)
        self.max_fl_fn = safe_fraction(self.max_fl_fn)

    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not file_value:
            return None
        file_value = file_value.split()
        if tag == 'Exif.CanonCs.Lens':
            long_focal, short_focal, focal_units = file_value
            if focal_units == '0':
                return None
            return cls(('{}/{}'.format(short_focal, focal_units),
                        '{}/{}'.format(long_focal, focal_units), 0, 0))
        return cls(file_value)

    def write(self, handler, tag):
        handler.set_string(tag, ' '.join(['{:d}/{:d}'.format(
            self[x].numerator, self[x].denominator) for x in self._keys]))

    def __str__(self):
        return ','.join(['{:g}'.format(float(self[x])) for x in self._keys])


class Thumbnail(MD_Dict):
    _keys = ('data', 'fmt', 'w', 'h')

    @classmethod
    def read(cls, handler, tag):
        if handler.is_xmp_tag(tag):
            data, fmt, w, h = handler.get_string(tag)
            if not all((data, fmt, w, h)):
                return None
            if not six.PY2:
                data = bytes(data, 'ASCII')
            data = codecs.decode(data, 'base64_codec')
            w = int(w)
            h = int(h)
        else:
            data = handler.get_exif_thumbnail()
            if not data:
                return None
            fmt = handler.get_tag_string(tag)
            fmt = ('TIFF', 'JPEG')[fmt == '6']
            w, h = None, None
        return cls((data, fmt, w, h))

    def write(self, handler, tag):
        if handler.is_xmp_tag(tag):
            data = self.data
            if self.fmt != 'JPEG':
                pixmap = QtGui.QPixmap()
                pixmap.loadFromData(data)
                buf = QtCore.QBuffer()
                buf.open(QtCore.QIODevice.WriteOnly)
                pixmap.save(buf, 'JPEG')
                data = buf.data().data()
                self.w = pixmap.width()
                self.h = pixmap.height()
            if not self.w or not self.h:
                pixmap = QtGui.QPixmap()
                pixmap.loadFromData(data)
                self.w = pixmap.width()
                self.h = pixmap.height()
            data = codecs.encode(data, 'base64_codec')
            handler.set_string(tag, (data, 'JPEG', str(self.w), str(self.h)))
        elif handler.get_supports_exif():
            handler.set_exif_thumbnail_from_buffer(self.data)

    def merge(self, info, tag, other):
        return self


class DateTime(MD_Dict):
    # store date and time with "precision" to store how much is valid
    # tz_offset is stored in minutes
    _keys = ('datetime', 'precision', 'tz_offset')

    def __init__(self, value):
        super(DateTime, self).__init__(value)
        if not self.datetime:
            # use a well known 'zero'
            self.datetime = datetime(1970, 1, 1)
        else:
            self.datetime = self.truncate_datetime(self.precision)
        if self.precision <= 3:
            self.tz_offset = None

    def truncate_datetime(self, precision):
        parts = [self.datetime.year, self.datetime.month, self.datetime.day,
                 self.datetime.hour, self.datetime.minute, self.datetime.second,
                 self.datetime.microsecond][:precision]
        parts.extend((1, 1, 1)[len(parts):])
        return datetime(*parts)

    @classmethod
    def from_ISO_8601(cls, date_string, time_string, tz_string):
        """Sufficiently general ISO 8601 parser.

        Inputs must be in "basic" format, i.e. no '-' or ':' separators.
        See https://en.wikipedia.org/wiki/ISO_8601

        """
        # parse tz_string
        if tz_string:
            tz_offset = (int(tz_string[1:3]) * 60) + int(tz_string[3:])
            if tz_string[0] == '-':
                tz_offset = -tz_offset
        else:
            tz_offset = None
        if time_string == '000000':
            # assume no time information
            time_string = ''
            tz_offset = None
        datetime_string = date_string + time_string[:13]
        precision = min((len(datetime_string) - 2) // 2, 7)
        if precision <= 0:
            return None
        fmt = ''.join(('%Y', '%m', '%d', '%H', '%M', '%S', '.%f')[:precision])
        return cls(
            (datetime.strptime(datetime_string, fmt), precision, tz_offset))

    def to_ISO_8601(self, fmt=('%Y', '-%m', '-%d', 'T%H', ':%M', ':%S', '.%f'),
                    precision=None, time_zone=True):
        if precision is None:
            precision = self.precision
        fmt = ''.join(fmt[:precision])
        datetime_string = self.datetime.strftime(fmt)
        if precision > 6:
            # truncate subsecond to 3 digits
            datetime_string = datetime_string[:-3]
        if precision > 3 and time_zone and self.tz_offset is not None:
            # add time zone
            minutes = self.tz_offset
            if minutes >= 0:
                datetime_string += '+'
            else:
                datetime_string += '-'
                minutes = -minutes
            datetime_string += '{:02d}:{:02d}'.format(
                minutes // 60, minutes % 60)
        return datetime_string

    # many quicktime movies use Apple's 1904 timestamp zero point
    qt_offset = (datetime(1970, 1, 1) - datetime(1904, 1, 1)).total_seconds()

    # Do something different with Xmp.video tags
    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not file_value:
            return None
        if handler.is_exif_tag(tag):
            return cls.from_exif(file_value)
        if handler.is_iptc_tag(tag):
            return cls.from_iptc(file_value)
        if tag.startswith('Xmp.video'):
            time_stamp = int(file_value)
            if time_stamp == 0:
                return None
            # assume date should be in range 1970 to 2034
            if time_stamp > cls.qt_offset:
                time_stamp -= cls.qt_offset
            return cls((datetime.utcfromtimestamp(time_stamp), 6, None))
        return cls.from_xmp(file_value)

    def write(self, handler, tag):
        if handler.is_exif_tag(tag):
            handler.set_string(tag, self.to_exif())
        elif handler.is_iptc_tag(tag):
            handler.set_string(tag, self.to_iptc())
        else:
            handler.set_string(tag, self.to_xmp())

    # Exif datetime is always full resolution and valid. Assume a time
    # of 00:00:00 is a None value though.
    @classmethod
    def from_exif(cls, file_value):
        datetime_string = file_value[0]
        if not datetime_string:
            return None
        # separate date & time and remove separators
        date_string = datetime_string[:10].replace(':', '')
        time_string = datetime_string[11:].replace(':', '')
        # append sub seconds
        if len(file_value) > 1:
            sub_sec_string = file_value[1]
            if sub_sec_string:
                time_string += '.' + sub_sec_string
        return cls.from_ISO_8601(date_string, time_string, '')

    def to_exif(self):
        datetime_string, sep, sub_sec_string = self.to_ISO_8601(
            fmt=('%Y', ':%m', ':%d', ' %H', ':%M', ':%S', '.%f'),
            time_zone=False).partition('.')
        # pad out any missing values
        #                   YYYY mm dd HH MM SS
        datetime_string += '0000:01:01 00:00:00'[len(datetime_string):]
        return datetime_string, sub_sec_string

    # IPTC date & time should have no separators and be 8 and 11 chars
    # respectively (time includes time zone offset). I suspect the exiv2
    # library is adding separators, but am not sure.

    # The date (and time?) can have missing values represented by 00
    # according to
    # https://de.wikipedia.org/wiki/IPTC-IIM-Standard#IPTC-Felder
    @classmethod
    def from_iptc(cls, file_value):
        date_string, time_string = file_value
        if not date_string:
            return None
        # remove separators (that shouldn't be there)
        date_string = date_string.replace('-', '')
        # remove missing values
        while len(date_string) > 4 and date_string[-2:] == '00':
            date_string = date_string[:-2]
        if date_string == '0000':
            return None
        # ignore time if date is not full precision
        if len(date_string) < 8:
            time_string = ''
        if time_string:
            # remove separators (that shouldn't be there)
            time_string = time_string.replace(':', '')
            # split off time zone
            tz_string = time_string[6:]
            time_string = time_string[:6]
        else:
            tz_string = ''
            time_string = ''
        return cls.from_ISO_8601(date_string, time_string, tz_string)

    def to_iptc(self):
        if self.precision <= 3:
            date_string = self.to_ISO_8601()
            #               YYYY mm dd
            date_string += '0000-00-00'[len(date_string):]
            time_string = None
        else:
            datetime_string = self.to_ISO_8601(precision=6)
            date_string = datetime_string[:10]
            time_string = datetime_string[11:]
        return date_string, time_string

    # XMP uses extended ISO 8601, but the time cannot be hours only. See
    # p75 of
    # https://partners.adobe.com/public/developer/en/xmp/sdk/XMPspecification.pdf
    # According to p71, when converting Exif values with no time zone,
    # local time zone should be assumed. However, the MWG guidelines say
    # this must not be assumed to be the time zone where the photo is
    # processed. It also says the XMP standard has been revised to make
    # time zone information optional.
    @classmethod
    def from_xmp(cls, file_value):
        date_string, sep, time_string = file_value.partition('T')
        if len(time_string) > 6 and time_string[-6] in ('+', '-'):
            tz_string = time_string[-6:]
            time_string = time_string[:-6]
        elif len(time_string) > 1 and time_string[-1] == 'Z':
            tz_string = '+00:00'
            time_string = time_string[:-1]
        else:
            tz_string = ''
        return cls.from_ISO_8601(
            date_string.replace('-', ''), time_string.replace(':', ''),
            tz_string.replace(':', ''))

    def to_xmp(self):
        precision = self.precision
        if precision == 4:
            precision = 5
        return self.to_ISO_8601(precision=precision)

    def __str__(self):
        return self.to_ISO_8601()

    def merge(self, info, tag, other):
        if other == self:
            return self
        merged = False
        if other.datetime != self.datetime:
            # if datetime values differ, choose the one with more precision
            if other.precision > self.precision:
                self.log_replaced(info, tag, other)
                return other
            if other.datetime != self.truncate_datetime(other.precision):
                self.log_ignored(info, tag, other)
                return self
        else:
            # some formats default to a higher precision than wanted
            if self.precision < 7 and other.precision < self.precision:
                self.precision = other.precision
                merged = True
        # don't trust IPTC time zone and Exif doesn't have time zone
        if (other.tz_offset not in (None, self.tz_offset) and
                MetadataHandler.is_xmp_tag(tag)):
            self.tz_offset = other.tz_offset
            merged = True
        if merged:
            self.log_merged(info, tag, other)
        return self


class MultiString(list, MD_Value):
    def __init__(self, value):
        if isinstance(value, six.string_types):
            value = value.split(';')
        value = filter(bool, [x.strip() for x in value])
        super(MultiString, self).__init__(value)

    @classmethod
    def read(cls, handler, tag):
        if handler.get_tag_type(tag) in ('String', 'XmpBag', 'XmpSeq'):
            file_value = handler.get_multiple(tag)
        else:
            file_value = handler.get_string(tag)
        if not file_value:
            return None
        if handler.get_tag_type(tag) == 'Byte':
            file_value = decode_UCS2(file_value)
        return cls(file_value)

    def write(self, handler, tag):
        if handler.is_exif_tag(tag):
            handler.set_string(tag, ';'.join(self))
        elif handler.get_tag_type(tag) in ('String', 'XmpBag', 'XmpSeq'):
            handler.set_multiple(tag, self)
        else:
            handler.set_string(tag, self)

    def __str__(self):
        return '; '.join(self)

    def merge(self, info, tag, other):
        merged = False
        for item in other:
            if item not in self:
                self.append(item)
                merged = True
        if merged:
            self.log_merged(info, tag, other)
        return self


class MD_String(six.text_type, MD_Value):
    @classmethod
    def read(cls, handler, tag):
        if handler.get_tag_type(tag) == 'LangAlt':
            file_value = handler.get_multiple(tag)
            if file_value:
                file_value = file_value[0]
        else:
            file_value = handler.get_string(tag)
        if file_value and handler.get_tag_type(tag) == 'Byte':
            file_value = decode_UCS2(file_value)
        if file_value:
            file_value = six.text_type(file_value).strip()
        if not file_value:
            return None
        return cls(file_value)

    def write(self, handler, tag):
        handler.set_string(tag, self)

    def merge(self, info, tag, other):
        if other in self:
            return self
        self.log_merged(info, tag, other)
        return MD_String(self + ' // ' + other)


class CharacterSet(MD_String):
    known_encodings = {
        'ascii'   : '\x1b(B',
        'latin_1' : '\x1b/A',
        'latin1'  : '\x1b.A',
        'utf_8'   : '\x1b%G',
        }

    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        for charset, encoding in cls.known_encodings.items():
            if encoding == file_value:
                return cls(charset)
        if file_value:
            logger.warning('Unknown character encoding "%s"', repr(file_value))
        return None

    def write(self, handler, tag):
        handler.set_string(tag, self.known_encodings[self])


class Software(MD_String):
    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if isinstance(file_value, list):
            program, version = file_value
            if not program:
                return None
            if version:
                program += ' v' + version
            return cls(program)
        if not file_value:
            return None
        return cls(file_value)

    def write(self, handler, tag):
        if handler.is_iptc_tag(tag):
            handler.set_string(tag, self.split(' v'))
        else:
            handler.set_string(tag, self)


class MD_Int(int, MD_Value):
    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not file_value:
            return None
        return cls(file_value)

    def write(self, handler, tag):
        handler.set_string(tag, six.text_type(self))

    def merge(self, info, tag, other):
        if self != other:
            self.log_ignored(info, tag, other)
        return self


class Timezone(MD_Int):
    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not file_value:
            return None
        if tag == 'Exif.Image.TimeZoneOffset':
            # convert hours to minutes
            return cls(int(file_value) * 60)
        return cls(file_value)


class Aperture(Fraction, MD_Value):
    # store FNumber and APEX aperture as fractions
    # only FNumber is presented to the user, either is computed if missing
    @classmethod
    def read(cls, handler, tag):
        file_value = handler.get_string(tag)
        if not any(file_value):
            return None
        f_number, apex = file_value
        if apex:
            apex = safe_fraction(apex)
        if not f_number:
            f_number = 2.0 ** (apex / 2.0)
        f_number = safe_fraction(f_number)
        self = cls(f_number)
        if apex:
            self.apex = apex
        return self

    def write(self, handler, tag):
        apex = getattr(self, 'apex', safe_fraction(math.log(self, 2) * 2.0))
        handler.set_string(tag, (
            '{:d}/{:d}'.format(self.numerator, self.denominator),
            '{:d}/{:d}'.format(apex.numerator, apex.denominator)))

    def __str__(self):
        return six.text_type(float(self))

    def merge(self, info, tag, other):
        if (min(other, self) / max(other, self)) < 0.95:
            self.log_ignored(info, tag, other)
        return self


# maximum length of Iptc data
_max_bytes = {
    'Iptc.Application2.Byline'           :   32,
    'Iptc.Application2.Caption'          : 2000,
    'Iptc.Application2.City'             :   32,
    'Iptc.Application2.Copyright'        :  128,
    'Iptc.Application2.CountryCode'      :    3,
    'Iptc.Application2.CountryName'      :   64,
    'Iptc.Application2.Headline'         :  256,
    'Iptc.Application2.Keywords'         :   64,
    'Iptc.Application2.ObjectName'       :   64,
    'Iptc.Application2.Program'          :   32,
    'Iptc.Application2.ProgramVersion'   :   10,
    'Iptc.Application2.ProvinceState'    :   32,
    'Iptc.Application2.SubLocation'      :   32,
    'Iptc.Envelope.CharacterSet'         :   32,
    }

# extra Xmp namespaces we may need
_extra_ns = {
    'Iptc4xmpExt': 'http://iptc.org/std/Iptc4xmpExt/2008-02-29/',
    }

class MetadataHandler(GExiv2.Metadata):
    def __init__(self, path):
        super(MetadataHandler, self).__init__()
        self._path = path
        # read metadata from file
        self.open_path(self._path)
        self._xmp_only = self.get_mime_type() in (
            'application/rdf+xml', 'application/postscript')
        # remove exiv2's synthesised non-XMP tags
        if self._xmp_only:
            self.clear_exif()
            self.clear_iptc()
        # make list of possible character encodings
        self._encodings = []
        for name in ('utf_8', 'latin_1'):
            self._encodings.append(codecs.lookup(name).name)
        char_set = locale.getdefaultlocale()[1]
        if char_set:
            try:
                name = codecs.lookup(char_set).name
                if name not in self._encodings:
                    self._encodings.append(name)
            except LookupError:
                pass
        # convert IPTC data to UTF-8
        if not self.has_iptc():
            return
        current_encoding = CharacterSet.read(self, 'Iptc.Envelope.CharacterSet')
        if current_encoding:
            if current_encoding == 'utf_8':
                return
            try:
                name = codecs.lookup(current_encoding).name
                if name not in self._encodings:
                    self._encodings.insert(0, name)
            except LookupError:
                pass
        for tag in self.get_iptc_tags():
            try:
                value_list = self.get_multiple(tag)
                self.set_multiple(tag, value_list)
            except Exception as ex:
                logger.exception(ex)

    def _decode_string(self, value):
        if not value:
            return value
        for encoding in self._encodings:
            try:
                return value.decode(encoding)
            except UnicodeDecodeError:
                continue
        return value.decode('utf_8', 'replace')

    def clear_value(self, tag):
        if isinstance(tag, tuple):
            for sub_tag in tag:
                self.clear_value(sub_tag)
            return
        if not self.has_tag(tag):
            return
        self.clear_tag(tag)
        if self.is_xmp_tag(tag) and '/' in tag:
            # attempt to remove XMP structure/container
            container, subtag = tag.split('/')
            bag = container.partition('[')[0]
            for t in self.get_xmp_tags():
                if t.startswith(container + '/'):
                    # bag is not empty
                    return
            if container != bag:
                self.clear_tag(container)
            self.clear_tag(bag)

    def get_string(self, tag):
        if isinstance(tag, tuple):
            return list(map(self.get_string, tag))
        try:
            result = self.get_tag_string(tag)
            if six.PY2:
                result = self._decode_string(result)
        except UnicodeDecodeError as ex:
            logger.error(str(ex))
            return None
        return result

    def get_multiple(self, tag):
        if isinstance(tag, tuple):
            return list(map(self.get_multiple, tag))
        try:
            result = self.get_tag_multiple(tag)
            if six.PY2:
                result = list(map(self._decode_string, result))
        except UnicodeDecodeError as ex:
            logger.error(str(ex))
            return []
        return result

    def set_string(self, tag, value):
        if isinstance(tag, tuple):
            for sub_tag, sub_value in zip(tag, value):
                self.set_string(sub_tag, sub_value)
            return
        if not value:
            self.clear_value(tag)
            return
        if tag in _max_bytes:
            value = value.encode('utf_8')[:_max_bytes[tag]]
            if not six.PY2:
                value = value.decode('utf_8', errors='ignore')
        elif six.PY2:
            value = value.encode('utf_8')
        if self.is_xmp_tag(tag) and '/' in tag:
            # create XMP structure/container
            container, subtag = tag.split('/')
            bag = container.partition('[')[0]
            ns_prefix = subtag.partition(':')[0]
            no_bag = True
            ns_defined = ns_prefix not in _extra_ns
            for t in self.get_xmp_tags():
                if t.startswith(bag):
                    # file already has container
                    no_bag = False
                if ns_prefix in t:
                    # file has correct namespace defined
                    ns_defined = True
            if no_bag:
                # create empty container
                if '[' in container:
                    self.set_xmp_tag_struct(bag, GExiv2.StructureType.ALT)
                else:
                    self.set_xmp_tag_struct(bag, GExiv2.StructureType.SEQ)
            if not ns_defined:
                # create some XMP data with the correct namespace
                data = '''<?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>
<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="XMP Core 4.4.0-Exiv2">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description
        xmlns:{}="{}">
    </rdf:Description>
  </rdf:RDF>
</x:xmpmeta>
<?xpacket end="w"?>'''.format(ns_prefix, _extra_ns[ns_prefix])
                if six.PY2:
                    data = data.decode('utf-8')
                # open the data to register the correct namespace
                md = GExiv2.Metadata()
                md.open_buf(data.encode('utf-8'))
        self.set_tag_string(tag, value)

    def set_multiple(self, tag, value):
        if isinstance(tag, tuple):
            for sub_tag, sub_value in zip(tag, value):
                self.set_multiple(sub_tag, sub_value)
            return
        if not value:
            self.clear_value(tag)
            return
        if self.is_iptc_tag(tag) and tag in _max_bytes:
            value = [x.encode('utf_8')[:_max_bytes[tag]] for x in value]
            if not six.PY2:
                value = [x.decode('utf_8') for x in value]
        elif six.PY2:
            value = [x.encode('utf_8') for x in value]
        self.set_tag_multiple(tag, value)

    @staticmethod
    def is_exif_tag(tag):
        if isinstance(tag, tuple):
            tag = tag[0]
        return GExiv2.Metadata.is_exif_tag(tag)

    @staticmethod
    def is_iptc_tag(tag):
        if isinstance(tag, tuple):
            tag = tag[0]
        return GExiv2.Metadata.is_iptc_tag(tag)

    @staticmethod
    def is_xmp_tag(tag):
        if isinstance(tag, tuple):
            tag = tag[0]
        return GExiv2.Metadata.is_xmp_tag(tag)

    def get_supports_exif(self):
        if self._xmp_only:
            return False
        return super(MetadataHandler, self).get_supports_exif()

    def has_iptc(self):
        if self._xmp_only:
            return False
        return super(MetadataHandler, self).has_iptc()

    def save(self, file_times):
        # don't try to save to unwritable formats
        if not (self.get_supports_xmp() or self.get_supports_exif()):
            return False
        # remove exiv2's synthesised non-XMP tags
        if self._xmp_only:
            self.clear_exif()
            self.clear_iptc()
        try:
            self.save_file(self._path)
            if file_times:
                os.utime(self._path, file_times)
        except Exception as ex:
            logger.exception(ex)
            return False
        return True

    def clone(self, other):
        # copy from other to self, ignoring thumbnails
        for tag in other.get_exif_tags():
            if tag.startswith('Exif.Thumbnail'):
                continue
            self.set_string(tag, other.get_string(tag))
        for tag in other.get_iptc_tags():
            self.set_multiple(tag, other.get_multiple(tag))
        for tag in other.get_xmp_tags():
            if tag.startswith('Xmp.xmp.Thumbnails'):
                continue
            if self.get_tag_type(tag) == 'XmpText':
                self.set_string(tag, other.get_string(tag))
            else:
                self.set_multiple(tag, other.get_multiple(tag))

    def get_all_tags(self):
        return self.get_exif_tags() + self.get_iptc_tags() + self.get_xmp_tags()

    def get_exif_thumbnail(self):
        thumb = super(MetadataHandler, self).get_exif_thumbnail()
        if using_pgi and isinstance(thumb, tuple):
            # get_exif_thumbnail returns (OK, data) tuple
            thumb = thumb[thumb[0]]
        if thumb:
            return bytearray(thumb)
        return None


class Metadata(QtCore.QObject):
    unsaved = QtCore.pyqtSignal(bool)

    # type of each Photini data field's data
    _data_type = {
        'aperture'       : Aperture,
        'camera_model'   : MD_String,
        'character_set'  : CharacterSet,
        'copyright'      : MD_String,
        'creator'        : MultiString,
        'date_digitised' : DateTime,
        'date_modified'  : DateTime,
        'date_taken'     : DateTime,
        'description'    : MD_String,
        'focal_length'   : FocalLength,
        'keywords'       : MultiString,
        'latlong'        : LatLon,
        'lens_make'      : MD_String,
        'lens_model'     : MD_String,
        'lens_serial'    : MD_String,
        'lens_spec'      : LensSpec,
        'location_shown' : Location,
        'location_taken' : Location,
        'orientation'    : MD_Int,
        'software'       : Software,
        'thumbnail'      : Thumbnail,
        'timezone'       : Timezone,
        'title'          : MD_String,
        }
    # Mapping of tags to Photini data fields Each field has a list of
    # (mode, tag) pairs, where tag can be a tuple of tags. The mode is a
    # string containing the read mode (RA (always), or RN (never)) and
    # write mode (WA (always), WX (if Exif not supported), W0 (clear the
    # tag), or WN (never).
    _tag_list = {
        'aperture'       : (('RA.WA', ('Exif.Photo.FNumber',
                                       'Exif.Photo.ApertureValue')),
                            ('RA.W0', ('Exif.Image.FNumber',
                                       'Exif.Image.ApertureValue')),
                            ('RA.WX', ('Xmp.exif.FNumber',
                                       'Xmp.exif.ApertureValue'))),
        'camera_model'   : (('RA.WN', 'Exif.Image.Model'),
                            ('RA.WN', 'Exif.Image.UniqueCameraModel'),
                            ('RA.WN', 'Xmp.video.Model')),
        'character_set'  : (('RA.WA', 'Iptc.Envelope.CharacterSet'),),
        'copyright'      : (('RA.WA', 'Exif.Image.Copyright'),
                            ('RA.WA', 'Xmp.dc.rights'),
                            ('RA.W0', 'Xmp.tiff.Copyright'),
                            ('RA.WA', 'Iptc.Application2.Copyright')),
        'creator'        : (('RA.WA', 'Exif.Image.Artist'),
                            ('RA.W0', 'Exif.Image.XPAuthor'),
                            ('RA.WA', 'Xmp.dc.creator'),
                            ('RA.W0', 'Xmp.tiff.Artist'),
                            ('RA.WA', 'Iptc.Application2.Byline')),
        'date_digitised' : (('RA.WA', ('Exif.Photo.DateTimeDigitized',
                                       'Exif.Photo.SubSecTimeDigitized')),
                            ('RA.WA', 'Xmp.xmp.CreateDate'),
                            ('RA.W0', 'Xmp.exif.DateTimeDigitized'),
                            ('RA.WN', 'Xmp.video.DateUTC'),
                            ('RA.WA', ('Iptc.Application2.DigitizationDate',
                                       'Iptc.Application2.DigitizationTime'))),
        'date_modified'  : (('RA.WA', ('Exif.Image.DateTime',
                                       'Exif.Photo.SubSecTime')),
                            ('RA.WA', 'Xmp.xmp.ModifyDate'),
                            ('RA.WN', 'Xmp.video.ModificationDate'),
                            ('RA.W0', 'Xmp.tiff.DateTime')),
        'date_taken'     : (('RA.WA', ('Exif.Photo.DateTimeOriginal',
                                       'Exif.Photo.SubSecTimeOriginal')),
                            ('RA.W0', ('Exif.Image.DateTimeOriginal',)),
                            ('RA.WA', 'Xmp.photoshop.DateCreated'),
                            ('RA.W0', 'Xmp.exif.DateTimeOriginal'),
                            ('RA.WN', 'Xmp.video.DateUTC'),
                            ('RA.WA', ('Iptc.Application2.DateCreated',
                                       'Iptc.Application2.TimeCreated'))),
        'description'    : (('RA.WA', 'Exif.Image.ImageDescription'),
                            ('RA.W0', 'Exif.Image.XPComment'),
                            ('RA.W0', 'Exif.Image.XPSubject'),
                            ('RA.WA', 'Xmp.dc.description'),
                            ('RA.W0', 'Xmp.tiff.ImageDescription'),
                            ('RA.WA', 'Iptc.Application2.Caption')),
        'focal_length'   : (('RA.WA', ('Exif.Photo.FocalLength',
                                       'Exif.Photo.FocalLengthIn35mmFilm')),
                            ('RA.W0', ('Exif.Image.FocalLength',)),
                            ('RA.WX', ('Xmp.exif.FocalLength',
                                       'Xmp.exif.FocalLengthIn35mmFilm'))),
        'keywords'       : (('RA.WA', 'Xmp.dc.subject'),
                            ('RA.WA', 'Iptc.Application2.Keywords'),
                            ('RA.W0', 'Exif.Image.XPKeywords')),
        'latlong'        : (('RA.WA', ('Exif.GPSInfo.GPSLatitude',
                                       'Exif.GPSInfo.GPSLatitudeRef',
                                       'Exif.GPSInfo.GPSLongitude',
                                       'Exif.GPSInfo.GPSLongitudeRef')),
                            ('RA.WX', ('Xmp.exif.GPSLatitude',
                                       'Xmp.exif.GPSLongitude')),
                            ('RA.WN', 'Xmp.video.GPSCoordinates')),
        'lens_make'      : (('RA.WA', 'Exif.Photo.LensMake'),
                            ('RA.WX', 'Xmp.exifEX.LensMake')),
        'lens_model'     : (('RA.WA', 'Exif.Photo.LensModel'),
                            ('RA.WX', 'Xmp.exifEX.LensModel'),
                            ('RA.W0', 'Exif.Canon.LensModel'),
                            ('RA.W0', 'Exif.OlympusEq.LensModel'),
                            ('RA.W0', 'Xmp.aux.Lens'),
                            ('RN.W0', 'Exif.CanonCs.LensType')),
        'lens_serial'    : (('RA.WA', 'Exif.Photo.LensSerialNumber'),
                            ('RA.WX', 'Xmp.exifEX.LensSerialNumber'),
                            ('RA.W0', 'Exif.OlympusEq.LensSerialNumber'),
                            ('RA.W0', 'Xmp.aux.SerialNumber')),
        'lens_spec'      : (('RA.WA', 'Exif.Photo.LensSpecification'),
                            ('RA.WX', 'Xmp.exifEX.LensSpecification'),
                            ('RA.W0', 'Exif.Image.LensInfo'),
                            ('RA.W0', 'Exif.CanonCs.Lens'),
                            ('RA.W0', 'Exif.Nikon3.Lens'),
                            ('RN.W0', 'Exif.CanonCs.ShortFocal'),
                            ('RN.W0', 'Exif.CanonCs.MaxAperture'),
                            ('RN.W0', 'Exif.CanonCs.MinAperture')),
        'location_shown' : (
            ('RA.WA', ('Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:Sublocation',
                       'Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:City',
                       'Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:ProvinceState',
                       'Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:CountryName',
                       'Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:CountryCode',
                       'Xmp.iptcExt.LocationShown[1]/Iptc4xmpExt:WorldRegion')),),
        'location_taken' : (
            ('RA.WA', ('Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:Sublocation',
                       'Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:City',
                       'Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:ProvinceState',
                       'Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:CountryName',
                       'Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:CountryCode',
                       'Xmp.iptcExt.LocationCreated[1]/Iptc4xmpExt:WorldRegion')),
            ('RA.WA', ('Xmp.iptc.Location',
                       'Xmp.photoshop.City',
                       'Xmp.photoshop.State',
                       'Xmp.photoshop.Country',
                       'Xmp.iptc.CountryCode')),
            ('RA.WA', ('Iptc.Application2.SubLocation',
                       'Iptc.Application2.City',
                       'Iptc.Application2.ProvinceState',
                       'Iptc.Application2.CountryName',
                       'Iptc.Application2.CountryCode'))),
        'orientation'    : (('RA.WA', 'Exif.Image.Orientation'),
                            ('RA.WX', 'Xmp.tiff.Orientation')),
        'software'       : (('RA.WA', 'Exif.Image.ProcessingSoftware'),
                            ('RA.WA', ('Iptc.Application2.Program',
                                       'Iptc.Application2.ProgramVersion'))),
        'thumbnail'      : (('RA.WA', 'Exif.Thumbnail.Compression'),
                            ('RA.WX', ('Xmp.xmp.Thumbnails[1]/xmpGImg:image',
                                       'Xmp.xmp.Thumbnails[1]/xmpGImg:format',
                                       'Xmp.xmp.Thumbnails[1]/xmpGImg:width',
                                       'Xmp.xmp.Thumbnails[1]/xmpGImg:height')),
                            ('RA.W0', ('Xmp.xmp.Thumbnails[1]/xapGImg:image',
                                       'Xmp.xmp.Thumbnails[1]/xapGImg:format',
                                       'Xmp.xmp.Thumbnails[1]/xapGImg:width',
                                       'Xmp.xmp.Thumbnails[1]/xapGImg:height'))),
        'timezone'       : (('RA.WN', 'Exif.Image.TimeZoneOffset'),
                            ('RA.WN', 'Exif.CanonTi.TimeZone'),
                            ('RA.WN', 'Exif.NikonWt.Timezone')),
        'title'          : (('RA.WA', 'Xmp.dc.title'),
                            ('RA.WA', 'Iptc.Application2.ObjectName'),
                            ('RA.W0', 'Exif.Image.XPTitle'),
                            ('RA.W0', 'Iptc.Application2.Headline')),
        }
    def __init__(self, path, *args, **kw):
        super(Metadata, self).__init__(*args, **kw)
        # create metadata handlers for image file and/or sidecar
        self._path = path
        self._sc_path = self._find_side_car(path)
        self._sc = None
        if self._sc_path:
            try:
                self._sc = MetadataHandler(self._sc_path)
            except Exception as ex:
                logger.exception(ex)
        self._if = None
        try:
            self._if = MetadataHandler(path)
        except GLib.Error:
            # expected if unrecognised file format
            pass
        except Exception as ex:
            logger.exception(ex)
        self.dirty = False

    def _find_side_car(self, path):
        for base in (os.path.splitext(path)[0], path):
            for ext in ('.xmp', '.XMP'):
                result = base + ext
                if os.path.exists(result):
                    return result
        return None

    def create_side_car(self):
        self._sc_path = self._path + '.xmp'
        with open(self._sc_path, 'w') as of:
            of.write('''<?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>
<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="XMP Core 4.4.0-Exiv2">
 <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <rdf:Description rdf:about=""
    xmlns:xmp="http://ns.adobe.com/xap/1.0/"
   xmp:CreatorTool="{}"/>
 </rdf:RDF>
</x:xmpmeta>
<?xpacket end="w"?>'''.format('Photini editor v' + __version__))
        try:
            self._sc = MetadataHandler(self._sc_path)
        except Exception as ex:
            logger.exception(ex)

    def save(self, if_mode, sc_mode, force_iptc, file_times):
        if not self.dirty:
            return
        if (sc_mode == 'always' or not self._if) and not self._sc:
            self.create_side_car()
        self.software = 'Photini editor v' + __version__
        self.character_set = 'utf_8'
        if self._sc:
            # workaround for bug in exiv2 xmp timestamp altering
            for name in ('date_digitised', 'date_modified', 'date_taken'):
                for mode, tag in self._tag_list[name]:
                    if mode == 'RA.WA':
                        self._sc.clear_value(tag)
            self._sc.save(file_times)
        for handler in (self._sc, self._if):
            if not handler:
                continue
            save_iptc = force_iptc or handler.has_iptc()
            for name in self._tag_list:
                value = getattr(self, name)
                for mode, tag in self._tag_list[name]:
                    if not save_iptc and handler.is_iptc_tag(tag):
                        continue
                    write_mode = mode.split('.')[1]
                    if write_mode == 'WN':
                        continue
                    if ((not value) or (write_mode == 'W0') or
                        (write_mode == 'WX' and handler.get_supports_exif())):
                        handler.clear_value(tag)
                    else:
                        value.write(handler, tag)
        if self._if and sc_mode == 'delete' and self._sc:
            self._if.clone(self._sc)
        OK = False
        if self._if and if_mode:
            OK = self._if.save(file_times)
            if OK:
                # check that data really was saved
                saved_tags = MetadataHandler(self._path).get_all_tags()
                for tag in self._if.get_all_tags():
                    if tag not in saved_tags:
                        logger.warning('tag not saved: %s', tag)
                        OK = False
            if not OK and not self._sc:
                # can't write to image so create side car
                self.save(False, 'always', force_iptc, file_times)
                return
        if sc_mode == 'delete' and self._sc and OK:
            os.unlink(self._sc_path)
            self._sc = None
        if self._sc:
            OK = self._sc.save(file_times)
        if OK:
            self.dirty = False
            self.unsaved.emit(self.dirty)

    def get_mime_type(self):
        if self._if:
            return self._if.get_mime_type()
        return None

    def copy(self, other):
        # copy from other to self, sidecar over-rides image
        if self._sc:
            if other._if:
                self._sc.clone(other._if)
            if other._sc:
                self._sc.clone(other._sc)
        if self._if:
            if other._if:
                self._if.clone(other._if)
            if other._sc:
                self._if.clone(other._sc)
        self.dirty = True
        self.unsaved.emit(self.dirty)

    def __getattr__(self, name):
        if name not in self._tag_list:
            raise AttributeError(
                "%s has no attribute %s" % (self.__class__, name))
        # read data values
        values = []
        for mode, tag in self._tag_list[name]:
            if mode.split('.')[0] == 'RN':
                continue
            new_value = None
            for handler in self._sc, self._if:
                if not handler:
                    continue
                try:
                    new_value = self._data_type[name].read(handler, tag)
                except Exception as ex:
                    logger.exception(ex)
                    continue
                if new_value:
                    values.append((tag, new_value))
                    break
        # choose result and merge in non-matching data so user can review it
        result = None
        if values:
            info = '{}({})'.format(os.path.basename(self._path), name)
            tag, result = values.pop(0)
            logger.debug('%s: set from %s', info, tag)
        for tag, value in values:
            result = result.merge(info, tag, value)
        # merge in camera timezone if needed
        if (result and name.startswith('date_') and
                            result.tz_offset is None and self.timezone):
            result.tz_offset = self.timezone
            logger.info('%s: merged camera timezone offset', info)
        # add value to object attributes so __getattr__ doesn't get
        # called again
        super(Metadata, self).__setattr__(name, result)
        return result

    def __setattr__(self, name, value):
        if name not in self._tag_list:
            return super(Metadata, self).__setattr__(name, value)
        if value in (None, '', [], {}):
            value = None
        elif not isinstance(value, self._data_type[name]):
            value = self._data_type[name](value)
            if not value:
                value = None
        if getattr(self, name) == value:
            return
        super(Metadata, self).__setattr__(name, value)
        if not self.dirty:
            self.dirty = True
            self.unsaved.emit(self.dirty)

    def changed(self):
        return self.dirty



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from collections import defaultdict
import logging
import os
import webbrowser

import pkg_resources
import six

from photini.imagelist import DRAG_MIMETYPE
from photini.pyqt import (
    Qt, QtCore, QtGui, QtWebChannel, QtWebEngineWidgets,
    QtWebKitWidgets, QtWidgets, set_symbol_font, SingleLineEdit, SquareButton)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate

if QtWebEngineWidgets:
    WebPageBase = QtWebEngineWidgets.QWebEnginePage
    WebViewBase = QtWebEngineWidgets.QWebEngineView
else:
    WebPageBase = QtWebKitWidgets.QWebPage
    WebViewBase = QtWebKitWidgets.QWebView

class WebPage(WebPageBase):
    def javaScriptConsoleMessage(self, msg, line, source):
        if msg.startswith("Consider using 'dppx' units instead of 'dpi'"):
            return
        logger.error('%s line %d: %s', source, line, msg)


class WebView(WebViewBase):
    drop_text = QtCore.pyqtSignal(int, int, six.text_type)

    def dragEnterEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(WebView, self).dragEnterEvent(event)
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(WebView, self).dragMoveEvent(event)

    def dropEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(WebView, self).dropEvent(event)
        text = event.mimeData().data(DRAG_MIMETYPE).data().decode('utf-8')
        if text:
            self.drop_text.emit(event.pos().x(), event.pos().y(), text)


class LocationWidgets(QtCore.QObject):
    new_value = QtCore.pyqtSignal(str, str)

    def __init__(self, *args, **kw):
        super(LocationWidgets, self).__init__(*args, **kw)
        self.members = {
            'sublocation'   : SingleLineEdit(),
            'city'          : SingleLineEdit(),
            'province_state': SingleLineEdit(),
            'country_name'  : SingleLineEdit(),
            'country_code'  : SingleLineEdit(),
            'world_region'  : SingleLineEdit(),
            }
        self.members['sublocation'].editingFinished.connect(self.new_sublocation)
        self.members['city'].editingFinished.connect(self.new_city)
        self.members['province_state'].editingFinished.connect(self.new_province_state)
        self.members['country_name'].editingFinished.connect(self.new_country_name)
        self.members['country_code'].editingFinished.connect(self.new_country_code)
        self.members['world_region'].editingFinished.connect(self.new_world_region)
        self.members['country_code'].setMaximumWidth(40)

    def __getitem__(self, key):
        return self.members[key]

    @QtCore.pyqtSlot()
    def new_sublocation(self):
        self.send_value('sublocation')

    @QtCore.pyqtSlot()
    def new_city(self):
        self.send_value('city')

    @QtCore.pyqtSlot()
    def new_province_state(self):
        self.send_value('province_state')

    @QtCore.pyqtSlot()
    def new_country_name(self):
        self.send_value('country_name')

    @QtCore.pyqtSlot()
    def new_country_code(self):
        self.send_value('country_code')

    @QtCore.pyqtSlot()
    def new_world_region(self):
        self.send_value('world_region')

    def send_value(self, key):
        self.new_value.emit(key, self.members[key].get_value())


class LocationInfo(QtWidgets.QWidget):
    def __init__(self, *args, **kw):
        super(LocationInfo, self).__init__(*args, **kw)
        layout = QtWidgets.QGridLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.members = {
            'taken': LocationWidgets(self),
            'shown': LocationWidgets(self)
            }
        self.swap = SquareButton(six.unichr(0x21c4))
        self.swap.setStyleSheet('QPushButton { font-size: 10px }')
        set_symbol_font(self.swap)
        layout.addWidget(self.swap, 0, 4)
        label = QtWidgets.QLabel(translate('PhotiniMap', 'camera'))
        layout.addWidget(label, 0, 1, 1, 2)
        label = QtWidgets.QLabel(translate('PhotiniMap', 'subject'))
        layout.addWidget(label, 0, 3)
        layout.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Street:')), 1, 0)
        layout.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'City:')), 2, 0)
        layout.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Province:')), 3, 0)
        layout.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Country:')), 4, 0)
        layout.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Region:')), 5, 0)
        for ts, col in (('taken', 1), ('shown', 3)):
            layout.addWidget(self.members[ts]['sublocation'], 1, col, 1, 2)
            layout.addWidget(self.members[ts]['city'], 2, col, 1, 2)
            layout.addWidget(self.members[ts]['province_state'], 3, col, 1, 2)
            layout.addWidget(self.members[ts]['country_name'], 4, col)
            layout.addWidget(self.members[ts]['country_code'], 4, col + 1)
            layout.addWidget(self.members[ts]['world_region'], 5, col, 1, 2)

    def __getitem__(self, key):
        return self.members[key]


class PhotiniMap(QtWidgets.QSplitter):
    def __init__(self, image_list, parent=None):
        super(PhotiniMap, self).__init__(parent)
        self.app = QtWidgets.QApplication.instance()
        self.config_store = self.app.config_store
        self.image_list = image_list
        self.script_dir = pkg_resources.resource_filename(
            'photini', 'data/' + self.__class__.__name__.lower() + '/')
        self.drag_icon = QtGui.QPixmap(
            os.path.join(self.script_dir, 'grey_marker.png'))
        self.search_string = None
        self.map_loaded = False
        self.marker_images = {}
        self.map_status = {}
        self.dropped_images = []
        self.setChildrenCollapsible(False)
        left_side = QtWidgets.QWidget()
        self.addWidget(left_side)
        self.grid = QtWidgets.QGridLayout()
        self.grid.setContentsMargins(0, 0, 0, 0)
        self.grid.setRowStretch(6, 1)
        self.grid.setColumnStretch(1, 1)
        left_side.setLayout(self.grid)
        # map
        self.map = WebView()
        self.map.setPage(WebPage(parent=self.map))
        if QtWebEngineWidgets:
            self.web_channel = QtWebChannel.QWebChannel()
            self.map.page().setWebChannel(self.web_channel)
            self.web_channel.registerObject('python', self)
        else:
            self.map.page().setLinkDelegationPolicy(
                QtWebKitWidgets.QWebPage.DelegateAllLinks)
            self.map.page().linkClicked.connect(self.link_clicked)
            self.map.page().mainFrame().javaScriptWindowObjectCleared.connect(
                self.java_script_window_object_cleared)
        self.map.setAcceptDrops(False)
        self.map.drop_text.connect(self.drop_text)
        self.addWidget(self.map)
        # search
        self.grid.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Search:')), 0, 0)
        self.edit_box = QtWidgets.QComboBox()
        self.edit_box.setMinimumWidth(200)
        self.edit_box.setEditable(True)
        self.edit_box.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
        self.edit_box.lineEdit().setPlaceholderText(
            translate('PhotiniMap', '<new search>'))
        self.edit_box.lineEdit().returnPressed.connect(self.search)
        self.edit_box.activated.connect(self.goto_search_result)
        self.clear_search()
        self.edit_box.setEnabled(False)
        self.grid.addWidget(self.edit_box, 0, 1, 1, 2)
        # latitude & longitude
        self.grid.addWidget(
            QtWidgets.QLabel(translate('PhotiniMap', 'Lat, long:')), 1, 0)
        self.coords = SingleLineEdit()
        self.coords.editingFinished.connect(self.new_coords)
        self.coords.setEnabled(False)
        self.grid.addWidget(self.coords, 1, 1)
        # convert lat/lng to location info
        self.auto_location = QtWidgets.QPushButton(
            translate('PhotiniMap', six.unichr(0x21e8) + ' address'))
        self.auto_location.setEnabled(False)
        self.auto_location.clicked.connect(self.get_address)
        self.grid.addWidget(self.auto_location, 1, 2)
        # location info
        self.location_info = LocationInfo()
        self.location_info['taken'].new_value.connect(self.new_location_taken)
        self.location_info['shown'].new_value.connect(self.new_location_shown)
        self.location_info.swap.clicked.connect(self.swap_locations)
        self.location_info.setEnabled(False)
        self.grid.addWidget(self.location_info, 3, 0, 1, 3)
        # load map button
        self.load_map = QtWidgets.QPushButton(translate('PhotiniMap', '\nLoad map\n'))
        self.load_map.clicked.connect(self.initialise)
        self.grid.addWidget(self.load_map, 7, 0, 1, 3)
        # other init
        self.image_list.image_list_changed.connect(self.image_list_changed)
        self.splitterMoved.connect(self.new_split)

    @QtCore.pyqtSlot(int, six.text_type)
    def log(self, level, message):
        logger.log(level, message)

    @QtCore.pyqtSlot(int, int)
    def new_split(self, pos, index):
        self.app.config_store.set('map', 'split', str(self.sizes()))

    @QtCore.pyqtSlot()
    def java_script_window_object_cleared(self):
        self.map.page().mainFrame().addToJavaScriptWindowObject("python", self)

    @QtCore.pyqtSlot(QtCore.QUrl)
    def link_clicked(self, url):
        if url.isLocalFile():
            url.setScheme('http')
        webbrowser.open_new(url.toString())

    @QtCore.pyqtSlot()
    def image_list_changed(self):
        self.redraw_markers()
        self.display_coords()
        self.display_location()
        self.see_selection()

    @QtCore.pyqtSlot()
    def initialise(self):
        page = '''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="initial-scale=1.0, user-scalable=no" />
    <style type="text/css">
      html, body {{ height: 100%; margin: 0; padding: 0 }}
      #mapDiv {{ position: relative; width: 100%; height: 100% }}
    </style>
{initialize}
{head}
    <script type="text/javascript" src="script.js"></script>
  </head>
  <body ondragstart="return false">
    <div id="mapDiv"></div>
    <script type="text/javascript">
      var initData = {{lat: {lat}, lng: {lng}, zoom: {zoom}}};
    </script>
{body}
  </body>
</html>
'''
        lat, lng = eval(self.config_store.get('map', 'centre', '(51.0, 0.0)'))
        zoom = eval(self.config_store.get('map', 'zoom', '11'))
        if QtWebEngineWidgets:
            initialize = '''
    <script type="text/javascript"
      src="qrc:///qtwebchannel/qwebchannel.js">
    </script>
    <script type="text/javascript">
      var python;
      function initialize()
      {
          new QWebChannel(qt.webChannelTransport, function (channel) {
              python = channel.objects.python;
              loadMap();
              });
      }
    </script>
'''
        else:
            initialize = '''
    <script type="text/javascript">
      function initialize()
      {
          loadMap();
      }
    </script>
'''
        page = page.format(lat=lat, lng=lng, zoom=zoom, initialize=initialize,
                           **self.get_page_elements())
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
        self.map.setHtml(page, QtCore.QUrl.fromLocalFile(self.script_dir))

    @QtCore.pyqtSlot()
    def initialize_finished(self):
        QtWidgets.QApplication.restoreOverrideCursor()
        self.map_loaded = True
        self.grid.removeWidget(self.load_map)
        self.load_map.setParent(None)
        show_terms = self.show_terms()
        self.grid.addLayout(show_terms, 7, 0, 1, 3)
        self.edit_box.setEnabled(True)
        self.map.setAcceptDrops(True)
        self.image_list.set_drag_to_map(self.drag_icon)
        self.redraw_markers()
        self.display_coords()

    def refresh(self):
        self.setSizes(
            eval(self.app.config_store.get('map', 'split', str(self.sizes()))))
        if not self.map_loaded:
            return
        lat, lng = eval(self.config_store.get('map', 'centre'))
        zoom = eval(self.config_store.get('map', 'zoom'))
        self.JavaScript('setView({!r},{!r},{:d})'.format(lat, lng, zoom))
        self.redraw_markers()
        self.image_list.set_drag_to_map(self.drag_icon)

    def do_not_close(self):
        return False

    @QtCore.pyqtSlot(QtCore.QVariant)
    def new_status(self, status):
        self.map_status = status
        self.config_store.set('map', 'centre', str(self.map_status['centre']))
        self.config_store.set('map', 'zoom', str(int(self.map_status['zoom'])))

    @QtCore.pyqtSlot(int, int, six.text_type)
    def drop_text(self, x, y, text):
        self.dropped_images = eval(text)
        self.JavaScript('markerDrop({:d},{:d})'.format(x, y))

    @QtCore.pyqtSlot(float, float)
    def marker_drop(self, lat, lng):
        for path in self.dropped_images:
            image = self.image_list.get_image(path)
            self._remove_image(image)
            self._set_metadata(image, lat, lng)
            self._add_image(image)
        self.dropped_images = []
        self.display_coords()
        self.see_selection()

    @QtCore.pyqtSlot()
    def new_coords(self):
        text = self.coords.get_value().strip()
        if not text:
            for image in self.image_list.get_selected_images():
                self._remove_image(image)
                image.metadata.latlong = None
            return
        try:
            lat, lng = map(float, text.split(','))
        except Exception:
            self.display_coords()
            return
        for image in self.image_list.get_selected_images():
            self._remove_image(image)
            self._set_metadata(image, lat, lng)
            self._add_image(image)
        self.display_coords()
        self.see_selection()

    def see_selection(self):
        locations = []
        for image in self.image_list.get_selected_images():
            latlong = image.metadata.latlong
            if not latlong:
                continue
            location = [latlong.lat, latlong.lon]
            if location not in locations:
                locations.append(location)
        if not locations:
            return
        self.JavaScript('fitPoints({})'.format(repr(locations)))

    @QtCore.pyqtSlot(six.text_type, six.text_type, six.text_type,
                     six.text_type, six.text_type, six.text_type)
    def set_location_taken(self, world_region, country_code, country_name,
                           province_state, city, sublocation):
        for image in self.image_list.get_selected_images():
            image.metadata.location_taken = (
                sublocation, city, province_state,
                country_name, country_code, world_region)
        self.display_location()

    @QtCore.pyqtSlot()
    def swap_locations(self):
        for image in self.image_list.get_selected_images():
            image.metadata.location_taken, image.metadata.location_shown = (
                image.metadata.location_shown, image.metadata.location_taken)
        self.display_location()

    @QtCore.pyqtSlot(six.text_type, six.text_type)
    def new_location_taken(self, key, value):
        self._new_location('location_taken', key, value)

    @QtCore.pyqtSlot(six.text_type, six.text_type)
    def new_location_shown(self, key, value):
        self._new_location('location_shown', key, value)

    def _new_location(self, taken_shown, key, value):
        for image in self.image_list.get_selected_images():
            location = getattr(image.metadata, taken_shown)
            if location:
                new_value = dict(location)
            else:
                new_value = dict.fromkeys((
                    'sublocation', 'city', 'province_state',
                    'country_name', 'country_code', 'world_region'))
            new_value[key] = value
            if not any(new_value.values()):
                new_value = None
            setattr(image.metadata, taken_shown, new_value)
        self.display_location()

    def display_coords(self):
        images = self.image_list.get_selected_images()
        if not images:
            self.coords.set_value(None)
            self.auto_location.setEnabled(False)
            return
        values = []
        for image in images:
            value = image.metadata.latlong
            if value not in values:
                values.append(value)
        if len(values) > 1:
            self.coords.set_multiple(choices=filter(None, values))
            self.auto_location.setEnabled(False)
        else:
            self.coords.set_value(values[0])
            self.auto_location.setEnabled(self.map_loaded and bool(values[0]))

    def display_location(self):
        images = self.image_list.get_selected_images()
        if not images:
            for widget_group in (self.location_info['taken'],
                                 self.location_info['shown']):
                for attr in widget_group.members:
                    widget_group[attr].set_value(None)
            return
        for taken_shown in 'taken', 'shown':
            widget_group = self.location_info[taken_shown]
            for attr in widget_group.members:
                values = []
                for image in images:
                    value = getattr(image.metadata, 'location_' + taken_shown)
                    if value:
                        value = value[attr]
                    if value not in values:
                        values.append(value)
                if len(values) > 1:
                    widget_group[attr].set_multiple(choices=filter(None, values))
                else:
                    widget_group[attr].set_value(values[0])

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        self.coords.setEnabled(bool(selection))
        self.location_info.setEnabled(bool(selection))
        for marker_id, images in self.marker_images.items():
            self.JavaScript('enableMarker({:d},{:d})'.format(
                marker_id, any([image.selected for image in images])))
        self.display_coords()
        self.display_location()
        self.see_selection()

    def redraw_markers(self):
        self.JavaScript('removeMarkers()')
        self.marker_images = {}
        for image in self.image_list.get_images():
            self._add_image(image)

    def _add_image(self, image):
        if not self.map_loaded:
            return
        latlong = image.metadata.latlong
        if not latlong:
            return
        for marker_id in self.marker_images:
            if self.marker_images[marker_id][0].metadata.latlong == latlong:
                self.marker_images[marker_id].append(image)
                if image.selected:
                    self.JavaScript(
                        'enableMarker({:d},{:d})'.format(marker_id, True))
                break
        else:
            for i in range(len(self.marker_images) + 2):
                marker_id = i
                if marker_id not in self.marker_images:
                    break
            self.marker_images[marker_id] = [image]
            self.JavaScript('addMarker({:d},{!r},{!r},{:d})'.format(
                marker_id, latlong.lat, latlong.lon,
                image.selected))

    def _remove_image(self, image):
        for marker_id in self.marker_images:
            if image in self.marker_images[marker_id]:
                break
        else:
            return
        self.marker_images[marker_id].remove(image)
        if self.marker_images[marker_id]:
            self.JavaScript('enableMarker({:d},{:d})'.format(
                marker_id,
                any([image.selected for image in self.marker_images[marker_id]])))
        else:
            self.JavaScript('delMarker({:d})'.format(marker_id))
            del self.marker_images[marker_id]

    @QtCore.pyqtSlot()
    def get_address(self):
        latlng = self.coords.get_value()
        self.JavaScript('reverseGeocode({})'.format(latlng))

    @QtCore.pyqtSlot()
    def search(self, search_string=None):
        if not search_string:
            search_string = self.edit_box.lineEdit().text()
            self.edit_box.clearEditText()
        if not search_string:
            return
        self.search_string = search_string
        self.clear_search()
        self.JavaScript('search("{}")'.format(search_string))

    def clear_search(self):
        self.edit_box.clear()
        self.edit_box.addItem('')
        if self.search_string:
            self.edit_box.addItem(translate('PhotiniMap', '<repeat search>'))

    @QtCore.pyqtSlot(float, float, float, float, six.text_type)
    def search_result(self, lat0, lng0, lat1, lng1, name):
        self.edit_box.addItem(name, (lat0, lng0, lat1, lng1))
        self.edit_box.showPopup()

    @QtCore.pyqtSlot(int)
    def goto_search_result(self, idx):
        self.edit_box.setCurrentIndex(0)
        self.edit_box.clearFocus()
        if idx == 0:
            return
        if self.search_string and idx == 1:
            # repeat search
            self.search(self.search_string)
            return
        view = self.edit_box.itemData(idx)
        self.JavaScript('adjustBounds({},{},{},{})'.format(*view))

    @QtCore.pyqtSlot(int)
    def marker_click(self, marker_id):
        self.image_list.select_images(self.marker_images[marker_id])

    @QtCore.pyqtSlot(float, float, int)
    def marker_drag(self, lat, lng, marker_id):
        for image in self.marker_images[marker_id]:
            self._set_metadata(image, lat, lng)
        self.display_coords()

    def _set_metadata(self, image, lat, lng):
        image.metadata.latlong = lat, lng

    def JavaScript(self, command):
        if self.map_loaded:
            if QtWebEngineWidgets:
                self.map.page().runJavaScript(command)
            else:
                self.map.page().mainFrame().evaluateJavaScript(command)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2015-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from collections import namedtuple
import re
import six

from photini.configstore import BaseConfigStore

# temporarily open config file to get any over-rides
config = BaseConfigStore('editor')
using_pyqt5 = config.get('pyqt', 'using_pyqt5', 'auto') != 'False'
using_qtwebengine = config.get('pyqt', 'using_qtwebengine', 'auto') != 'False'

if using_pyqt5:
    try:
        from PyQt5 import QtCore
    except ImportError:
        using_pyqt5 = False

if using_pyqt5:
    from PyQt5 import QtGui, QtWidgets
    from PyQt5.QtCore import Qt
    from PyQt5.QtNetwork import QNetworkProxy
    if using_qtwebengine:
        try:
            from PyQt5 import QtWebChannel, QtWebEngineWidgets
        except ImportError:
            using_qtwebengine = False
    if using_qtwebengine:
        QtWebKit = None
        QtWebKitWidgets = None
    else:
        from PyQt5 import QtWebKit, QtWebKitWidgets
        QtWebChannel = None
        QtWebEngineWidgets = None
    import sip
    sip.setdestroyonexit(True)
else:
    import sip
    sip.setapi('QString', 2)
    sip.setapi('QVariant', 2)
    from PyQt4 import QtCore, QtGui, QtWebKit
    using_qtwebengine = False
    QtWidgets = QtGui
    QtWebKitWidgets = QtWebKit
    QtWebChannel = None
    QtWebEngineWidgets = None
    from PyQt4.QtCore import Qt
    from PyQt4.QtNetwork import QNetworkProxy

style = config.get('pyqt', 'style')
if style:
    QtWidgets.QApplication.setStyle(style)
config.save()
del config, style

qt_version_info = namedtuple(
    'qt_version_info', ('major', 'minor', 'micro'))._make(
        map(int, QtCore.QT_VERSION_STR.split('.')))

def image_types():
    result = [
        'jpeg', 'jpg', 'exv', 'cr2', 'crw', 'mrw', 'tiff', 'tif', 'dng',
        'nef', 'pef', 'arw', 'rw2', 'sr2', 'srw', 'orf', 'png', 'pgf',
        'raf', 'eps', 'gif', 'psd', 'tga', 'bmp', 'jp2', 'pnm'
        ]
    for fmt in QtGui.QImageReader.supportedImageFormats():
        ext = fmt.data().decode('utf_8').lower()
        if ext not in result:
            result.append(ext)
    for ext in ('ico', 'xcf'):
        if ext in result:
            result.remove(ext)
    return result

def video_types():
    return ['3gp', 'avi', 'mp4', 'mpeg', 'mpg', 'mov', 'mts', 'qt', 'wmv']

def multiple():
    return QtCore.QCoreApplication.translate('Multiple', '<multiple>')

def multiple_values():
    return QtCore.QCoreApplication.translate('Multiple', '<multiple values>')

def set_symbol_font(widget):
    widget.setFont(QtGui.QFont('DejaVu Sans'))
    if widget.fontInfo().family().lower() != 'dejavu sans':
        # probably on Windows, try a different font
        widget.setFont(QtGui.QFont("Segoe UI Symbol"))

class Busy(object):
    @staticmethod
    def start():
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)

    @staticmethod
    def stop():
        QtWidgets.QApplication.restoreOverrideCursor()

    def __enter__(self):
        Busy.start()
        return self

    def __exit__(self, type, value, traceback):
        Busy.stop()


class SpellingHighlighter(QtGui.QSyntaxHighlighter):
    words = re.compile(r"\w+([-'’]\w+)*", flags=re.IGNORECASE | re.UNICODE)

    def __init__(self, *arg, **kw):
        super(SpellingHighlighter, self).__init__(*arg, **kw)
        self.spell_check = QtWidgets.QApplication.instance().spell_check
        self.spell_check.new_dict.connect(self.rehighlight)

    def findWords(self, text):
        for word in self.words.finditer(text):
            yield word.group(), word.start(), word.end()

    def highlightBlock(self, text):
        if not (text and self.spell_check.enabled and self.spell_check.dict):
            return
        formatter = QtGui.QTextCharFormat()
        formatter.setUnderlineColor(Qt.red)
        formatter.setUnderlineStyle(QtGui.QTextCharFormat.SpellCheckUnderline)
        for word, start, end in self.findWords(text):
            if not self.spell_check.dict.check(word):
                self.setFormat(start, end - start, formatter)

    def suggestions(self, word):
        if not (self.spell_check.enabled and self.spell_check.dict):
            return []
        if self.spell_check.dict.check(word):
            return []
        return self.spell_check.dict.suggest(word)


class MultiLineEdit(QtWidgets.QPlainTextEdit):
    editingFinished = QtCore.pyqtSignal()

    def __init__(self, spell_check=False, *arg, **kw):
        super(MultiLineEdit, self).__init__(*arg, **kw)
        self.multiple_values = multiple_values()
        self.setTabChangesFocus(True)
        self._is_multiple = False
        if spell_check:
            self.spell_check = SpellingHighlighter(self.document())
        else:
            self.spell_check = None

    def focusOutEvent(self, event):
        if not self._is_multiple:
            self.editingFinished.emit()
        super(MultiLineEdit, self).focusOutEvent(event)

    def keyPressEvent(self, event):
        if self._is_multiple:
            self._is_multiple = False
            if qt_version_info >= (5, 3):
                self.setPlaceholderText('')
        super(MultiLineEdit, self).keyPressEvent(event)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        suggestion_group = QtWidgets.QActionGroup(menu)
        if self._is_multiple:
            if self.choices:
                sep = menu.insertSeparator(menu.actions()[0])
                fm = menu.fontMetrics()
                for suggestion in self.choices:
                    label = six.text_type(suggestion).replace('\n', ' ')
                    label = fm.elidedText(label, Qt.ElideMiddle, self.width())
                    action = QtWidgets.QAction(label, suggestion_group)
                    action.setData(six.text_type(suggestion))
                    menu.insertAction(sep, action)
        elif self.spell_check:
            cursor = self.cursorForPosition(event.pos())
            block_pos = cursor.block().position()
            for word, start, end in self.spell_check.findWords(cursor.block().text()):
                if start > cursor.positionInBlock():
                    break
                if end <= cursor.positionInBlock():
                    continue
                cursor.setPosition(block_pos + start)
                cursor.setPosition(block_pos + end, QtGui.QTextCursor.KeepAnchor)
                break
            word = cursor.selectedText()
            if word:
                suggestions = self.spell_check.suggestions(word)
                if suggestions:
                    sep = menu.insertSeparator(menu.actions()[0])
                    for suggestion in suggestions:
                        action = QtWidgets.QAction(suggestion, suggestion_group)
                        menu.insertAction(sep, action)
        action = menu.exec_(event.globalPos())
        if action and action.actionGroup() == suggestion_group:
            if self._is_multiple:
                self.set_value(action.data())
            else:
                cursor.setPosition(block_pos + start)
                cursor.setPosition(block_pos + end, QtGui.QTextCursor.KeepAnchor)
                cursor.insertText(action.iconText())

    def set_value(self, value):
        self._is_multiple = False
        if not value:
            self.clear()
            if qt_version_info >= (5, 3):
                self.setPlaceholderText('')
        else:
            self.setPlainText(six.text_type(value))

    def get_value(self):
        value = self.toPlainText()
        if qt_version_info < (5, 3) and value == self.multiple_values:
            return ''
        return value

    def set_multiple(self, choices=[]):
        self._is_multiple = True
        self.choices = list(choices)
        if qt_version_info >= (5, 3):
            self.setPlaceholderText(self.multiple_values)
            self.clear()
        else:
            self.setPlainText(self.multiple_values)

    def is_multiple(self):
        return self._is_multiple and not bool(self.get_value())


class SingleLineEdit(MultiLineEdit):
    def __init__(self, *arg, **kw):
        super(SingleLineEdit, self).__init__(*arg, **kw)
        self.setFixedHeight(
            self.fontMetrics().lineSpacing() + 8 + (self.frameWidth() * 2))
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            event.ignore()
            return
        super(SingleLineEdit, self).keyPressEvent(event)

    def insertFromMimeData(self, source):
        self.insertPlainText(source.text().replace('\n', ' '))


class SquareButton(QtWidgets.QPushButton):
    def sizeHint(self):
        size = super(SquareButton, self).sizeHint()
        size.setHeight(size.height() - 4)
        size.setWidth(size.height())
        return size


class StartStopButton(QtWidgets.QPushButton):
    click_start = QtCore.pyqtSignal()
    click_stop = QtCore.pyqtSignal()

    def __init__(self, start_text, stop_text, *arg, **kw):
        super(StartStopButton, self).__init__(*arg, **kw)
        self.start_text = start_text
        self.stop_text = stop_text
        self.setCheckable(True)
        self.toggled.connect(self.toggle_text)
        self.clicked.connect(self.do_clicked)
        # get a size big enough for either text
        self.setText(self.stop_text)
        stop_size = super(StartStopButton, self).sizeHint()
        self.setText(self.start_text)
        start_size = super(StartStopButton, self).sizeHint()
        self.minimum_size = stop_size.expandedTo(start_size)

    def sizeHint(self):
        return self.minimum_size

    @QtCore.pyqtSlot(bool)
    def toggle_text(self, checked):
        if checked:
            self.setText(self.stop_text)
        else:
            self.setText(self.start_text)

    @QtCore.pyqtSlot(bool)
    def do_clicked(self, checked):
        if checked:
            self.click_start.emit()
        else:
            self.click_stop.emit()



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import six
from datetime import datetime
import imghdr
import logging
import mimetypes
import os
from six import BytesIO
from six.moves.urllib.parse import unquote
import webbrowser

try:
    import cv2
    import numpy as np
except ImportError:
    cv2 = None
try:
    import PIL.Image as PIL
except ImportError:
    PIL = None

from photini.metadata import Metadata
from photini.pyqt import (Busy, image_types, Qt, QtCore, QtGui, QtWidgets,
                          qt_version_info, set_symbol_font, video_types)

logger = logging.getLogger(__name__)
DRAG_MIMETYPE = 'application/x-photini-image'

class Image(QtWidgets.QFrame):
    def __init__(self, path, image_list, thumb_size=80, *arg, **kw):
        super(Image, self).__init__(*arg, **kw)
        self.path = path
        self.image_list = image_list
        self.name, ext = os.path.splitext(os.path.basename(self.path))
        self.selected = False
        self.thumb_size = thumb_size
        # read metadata
        self.metadata = Metadata(self.path)
        self.metadata.unsaved.connect(self.show_status)
        self.file_times = (os.path.getatime(self.path),
                           os.path.getmtime(self.path))
        # set file type
        self.file_type = self.metadata.get_mime_type()
        if not self.file_type:
            self.file_type = mimetypes.guess_type(self.path)[0]
        if not self.file_type:
            self.file_type = imghdr.what(self.path)
            if self.file_type:
                self.file_type = 'image/' + self.file_type
        # anything not recognised is assumed to be 'raw'
        if not self.file_type:
            self.file_type = 'image/raw'
        # sub widgets
        layout = QtWidgets.QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(3, 3, 3, 3)
        self.setLayout(layout)
        self.setToolTip(self.path)
        # label to display image
        self.image = QtWidgets.QLabel()
        self.image.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        layout.addWidget(self.image, 0, 0, 1, 2)
        # label to display file name
        self.label = QtWidgets.QLabel()
        self.label.setAlignment(Qt.AlignRight)
        self.label.setStyleSheet("QLabel { font-size: 12px }")
        layout.addWidget(self.label, 1, 1)
        # label to display status
        self.status = QtWidgets.QLabel()
        self.status.setAlignment(Qt.AlignLeft)
        self.status.setStyleSheet("QLabel { font-size: 12px }")
        set_symbol_font(self.status)
        layout.addWidget(self.status, 1, 0)
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Plain)
        self.setObjectName("thumbnail")
        self.set_selected(False)
        self.show_status(False)
        self._set_thumb_size(self.thumb_size)

    def reload_metadata(self):
        self.metadata = Metadata(self.path)
        self.metadata.unsaved.connect(self.show_status)
        self.show_status(False)
        self.load_thumbnail()
        self.image_list.emit_selection()

    def save_metadata(self):
        self.image_list.save_files(images=[self])

    def get_video_frame(self):
        if not cv2:
            return
        video = cv2.VideoCapture(self.path)
        if not video.isOpened():
            return
        OK, cv_image = video.read()
        if not OK:
            return
        height, width, channel = cv_image.shape
        fmt = QtGui.QImage.Format_RGB888
        # need to pad to 4 pixel multiple
        new_width = width - (width % -4)
        if channel == 4:
            # assume BGRA
            fmt = QtGui.QImage.Format_ARGB32
            np_image = np.empty((height, new_width, channel), dtype=np.uint8)
            np_image[ : , :width, 0] = cv_image[ : , : , 3]
            np_image[ : , :width, 1] = cv_image[ : , : , 2]
            np_image[ : , :width, 2] = cv_image[ : , : , 1]
            np_image[ : , :width, 3] = cv_image[ : , : , 0]
        elif channel == 3:
            # assume BGR
            np_image = np.empty((height, new_width, channel), dtype=np.uint8)
            np_image[ : , :width, 0] = cv_image[ : , : , 2]
            np_image[ : , :width, 1] = cv_image[ : , : , 1]
            np_image[ : , :width, 2] = cv_image[ : , : , 0]
        elif channel == 1:
            # assume Y
            channel = 3
            np_image = np.empty((height, new_width, channel), dtype=np.uint8)
            np_image[ : , :width, 0] = cv_image[ : , : , 0]
            np_image[ : , :width, 1] = cv_image[ : , : , 0]
            np_image[ : , :width, 2] = cv_image[ : , : , 0]
        else:
            return
        bpl = new_width * channel
        qt_im = QtGui.QImage(np_image.data, width, height, bpl, fmt)
        # attach np_image so it isn't deleted until qt_im is
        qt_im._data = np_image
        return qt_im

    def regenerate_thumbnail(self):
        with Busy():
            # get Qt image first
            qt_im = QtGui.QImage(self.path)
            if self.file_type.startswith('video') and qt_im.isNull():
                # use OpenCV to read first frame
                qt_im = self.get_video_frame()
            if not qt_im or qt_im.isNull():
                logger.error('Cannot read image data from %s', self.path)
                return
            # reorient if required
            if self.file_type == 'image/x-canon-cr2':
                orientation = self.metadata.orientation
                if orientation and orientation > 1:
                    # need to unrotate and or unreflect image
                    transform = QtGui.QTransform()
                    if orientation in (3, 4):
                        transform = transform.rotate(180.0)
                    elif orientation in (5, 6):
                        transform = transform.rotate(-90.0)
                    elif orientation in (7, 8):
                        transform = transform.rotate(90.0)
                    if orientation in (2, 4, 5, 7):
                        transform = transform.scale(-1.0, 1.0)
                    qt_im = qt_im.transformed(transform)
            # DCF spec says thumbnail must be 160 x 120 so pad picture
            # to 4:3 aspect ratio
            w = qt_im.width()
            h = qt_im.height()
            new_h = int(0.5 + (float(w * 3) / 4.0))
            new_w = int(0.5 + (float(h * 4) / 3.0))
            if new_h > h:
                pad = (new_h - h) // 2
                qt_im = qt_im.copy(0, -pad, w, new_h)
            elif new_w > w:
                pad = (new_w - w) // 2
                qt_im = qt_im.copy(-pad, 0, new_w, h)
            fmt = 'JPEG'
            if PIL:
                # convert Qt image to PIL image
                buf = QtCore.QBuffer()
                buf.open(QtCore.QIODevice.WriteOnly)
                qt_im.save(buf, 'PPM')
                data = BytesIO(buf.data().data())
                pil_im = PIL.open(data)
                # scale PIL image
                pil_im = pil_im.resize((160, 120), PIL.ANTIALIAS)
                w, h = pil_im.size
                # save image to memory
                data = BytesIO()
                pil_im.save(data, fmt)
                data = data.getvalue()
            else:
                # scale Qt image - not as good quality as PIL
                qt_im = qt_im.scaled(
                    160, 120, Qt.IgnoreAspectRatio, Qt.SmoothTransformation)
                w = qt_im.width()
                h = qt_im.height()
                # save image to memory
                buf = QtCore.QBuffer()
                buf.open(QtCore.QIODevice.WriteOnly)
                qt_im.save(buf, fmt)
                data = buf.data().data()
            # set thumbnail
            self.metadata.thumbnail = data, fmt, w, h
            # reload thumbnail
            self.load_thumbnail()

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        menu.addAction(self.tr('Reload metadata'), self.reload_metadata)
        menu.addAction(self.tr('Save metadata'), self.save_metadata)
        menu.addAction(self.tr('Regenerate thumbnail'), self.regenerate_thumbnail)
        action = menu.exec_(event.globalPos())

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.drag_start_pos = event.pos()
        if event.modifiers() == Qt.ControlModifier:
            self.image_list.select_image(self, multiple_selection=True)
        elif event.modifiers() == Qt.ShiftModifier:
            self.image_list.select_image(self, extend_selection=True)
        elif not self.get_selected():
            # don't clear selection in case we're about to drag
            self.image_list.select_image(self)

    def mouseReleaseEvent(self, event):
        if event.modifiers() not in (Qt.ControlModifier, Qt.ShiftModifier):
            # clear any multiple selection
            self.image_list.select_image(self)

    def mouseMoveEvent(self, event):
        if not self.image_list.drag_icon:
            return
        if ((event.pos() - self.drag_start_pos).manhattanLength() <
                                    QtWidgets.QApplication.startDragDistance()):
            return
        paths = []
        for image in self.image_list.get_selected_images():
            paths.append(image.path)
        if not paths:
            return
        drag = QtGui.QDrag(self)
        # construct icon
        count = min(len(paths), 8)
        src_icon = self.image_list.drag_icon
        src_w = src_icon.width()
        src_h = src_icon.height()
        margin = (count - 1) * 4
        if count == 1:
            icon = src_icon
        else:
            icon = QtGui.QPixmap(src_w + margin, src_h + margin)
            icon.fill(Qt.transparent)
            with QtGui.QPainter(icon) as paint:
                for i in range(count):
                    paint.drawPixmap(
                        QtCore.QPoint(margin - (i * 4), i * 4), src_icon)
        drag.setPixmap(icon)
        if src_h == src_w:
            # round marker used in Bing maps version 8
            drag.setHotSpot(QtCore.QPoint(src_w // 2, (src_h // 2) + margin))
        else:
            drag.setHotSpot(QtCore.QPoint(src_w // 2, src_h + margin))
        mimeData = QtCore.QMimeData()
        mimeData.setData(DRAG_MIMETYPE, repr(paths).encode('utf-8'))
        drag.setMimeData(mimeData)
        dropAction = drag.exec_(Qt.CopyAction)

    def mouseDoubleClickEvent(self, event):
        webbrowser.open(self.path)

    @QtCore.pyqtSlot(bool)
    def show_status(self, changed):
        status = ''
        # set 'geotagged' status
        if self.metadata.latlong:
            status += six.unichr(0x2690)
        # set 'unsaved' status
        if changed:
            status += six.unichr(0x26A1)
        self.status.setText(status)
        self._elide_name()
        if changed:
            self.image_list.new_metadata.emit(True)

    def _elide_name(self):
        self.status.adjustSize()
        elided_name = self.label.fontMetrics().elidedText(
            self.name, Qt.ElideLeft, self.thumb_size - self.status.width())
        self.label.setText(elided_name)

    def _set_thumb_size(self, thumb_size):
        self.thumb_size = thumb_size
        self.image.setFixedSize(self.thumb_size, self.thumb_size)
        self._elide_name()

    def set_thumb_size(self, thumb_size):
        self._set_thumb_size(thumb_size)
        self.load_thumbnail()

    def load_thumbnail(self):
        pixmap = QtGui.QPixmap()
        thumb = self.metadata.thumbnail
        if thumb:
            pixmap.loadFromData(thumb.data)
        if pixmap.isNull():
            self.image.setText(self.tr('No\nthumbnail\nin file'))
            return
        orientation = self.metadata.orientation
        if orientation and orientation > 1:
            # need to rotate and or reflect image
            transform = QtGui.QTransform()
            if orientation in (3, 4):
                transform = transform.rotate(180.0)
            elif orientation in (5, 6):
                transform = transform.rotate(90.0)
            elif orientation in (7, 8):
                transform = transform.rotate(-90.0)
            if orientation in (2, 4, 5, 7):
                transform = transform.scale(-1.0, 1.0)
            pixmap = pixmap.transformed(transform)
        self.image.setPixmap(
            pixmap.scaled(self.thumb_size, self.thumb_size,
                          Qt.KeepAspectRatio, Qt.SmoothTransformation))

    def set_selected(self, value):
        self.selected = value
        if self.selected:
            self.setStyleSheet("#thumbnail {border: 2px solid red}")
        else:
            self.setStyleSheet("#thumbnail {border: 2px solid grey}")

    def get_selected(self):
        return self.selected


class ScrollArea(QtWidgets.QScrollArea):
    dropped_images = QtCore.pyqtSignal(list)

    def __init__(self, parent=None):
        super(ScrollArea, self).__init__(parent)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setWidgetResizable(True)
        self.setAcceptDrops(True)

    def dropEvent(self, event):
        file_list = []
        for uri in event.mimeData().urls():
            file_list.append(uri.toLocalFile())
        if file_list:
            self.dropped_images.emit(file_list)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('text/uri-list'):
            event.acceptProposedAction()


class FlowLayout(QtWidgets.QLayout):
    """Left to right, top to bottom, fill available space Qt layout.

    Python implementation, based on C++ example at
    http://doc.qt.io/qt-4.8/qt-layouts-flowlayout-example.html

    """
    def __init__(self, *arg, **kw):
        super(FlowLayout, self).__init__(*arg, **kw)
        self.item_list = []

    def addItem(self, item):
        self.item_list.append(item)

    def horizontalSpacing(self):
        return 0

    def verticalSpacing(self):
        return 0

    def count(self):
        return len(self.item_list)

    def itemAt(self, idx):
        if idx < 0 or idx >= len(self.item_list):
            return None
        return self.item_list[idx]

    def takeAt(self, idx):
        if idx < 0 or idx >= len(self.item_list):
            return None
        return self.item_list.pop(idx)

    def expandingDirections(self):
        return 0

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self._do_layout(QtCore.QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self._do_layout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QtCore.QSize()
        for item in self.item_list:
            size = size.expandedTo(item.minimumSize())
        left, top, right, bottom = self.getContentsMargins()
        size += QtCore.QSize(left + right, top + bottom)
        return size

    def _do_layout(self, rect, test_only):
        left, top, right, bottom = self.getContentsMargins()
        effective_rect = rect.adjusted(left, top, -right, -bottom)
        x = effective_rect.x()
        y = effective_rect.y()
        row_height = 0
        for item in self.item_list:
            item_size = item.sizeHint()
            if x + item_size.width() > effective_rect.right() and row_height > 0:
                x = effective_rect.x()
                y += row_height
                row_height = 0
            if not test_only:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item_size))
            x += item_size.width()
            row_height = max(row_height, item_size.height())
        return y + row_height - rect.y() + bottom


class ImageList(QtWidgets.QWidget):
    image_list_changed = QtCore.pyqtSignal()
    new_metadata = QtCore.pyqtSignal(bool)
    selection_changed = QtCore.pyqtSignal(list)
    sort_order_changed = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super(ImageList, self).__init__(parent)
        self.app = QtWidgets.QApplication.instance()
        self.drag_icon = None
        self.images = []
        self.last_selected = None
        self.selection_anchor = None
        self.thumb_size = int(
            self.app.config_store.get('controls', 'thumb_size', '80'))
        layout = QtWidgets.QGridLayout()
        layout.setSpacing(0)
        layout.setRowStretch(0, 1)
        layout.setColumnStretch(3, 1)
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        # thumbnail display
        self.scroll_area = ScrollArea()
        self.scroll_area.dropped_images.connect(self.open_file_list)
        layout.addWidget(self.scroll_area, 0, 0, 1, 6)
        self.thumbnails = QtWidgets.QWidget()
        self.thumbnails.setLayout(FlowLayout())
        self.scroll_area.setWidget(self.thumbnails)
        QtWidgets.QShortcut(QtGui.QKeySequence.MoveToPreviousChar,
                        self.scroll_area, self.move_to_prev_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.MoveToNextChar,
                        self.scroll_area, self.move_to_next_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.MoveToStartOfLine,
                        self.scroll_area, self.move_to_first_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.MoveToEndOfLine,
                        self.scroll_area, self.move_to_last_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.SelectPreviousChar,
                        self.scroll_area, self.select_prev_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.SelectNextChar,
                        self.scroll_area, self.select_next_thumb)
        QtWidgets.QShortcut(QtGui.QKeySequence.SelectAll,
                        self.scroll_area, self.select_all)
        # sort key selector
        layout.addWidget(QtWidgets.QLabel(self.tr('sort by: ')), 1, 0)
        self.sort_name = QtWidgets.QRadioButton(self.tr('file name'))
        self.sort_name.clicked.connect(self._new_sort_order)
        layout.addWidget(self.sort_name, 1, 1)
        self.sort_date = QtWidgets.QRadioButton(self.tr('date taken'))
        layout.addWidget(self.sort_date, 1, 2)
        self.sort_date.clicked.connect(self._new_sort_order)
        if eval(self.app.config_store.get('controls', 'sort_date', 'False')):
            self.sort_date.setChecked(True)
        else:
            self.sort_name.setChecked(True)
        # size selector
        layout.addWidget(QtWidgets.QLabel(self.tr('thumbnail size: ')), 1, 4)
        self.size_slider = QtWidgets.QSlider(Qt.Horizontal)
        self.size_slider.setTracking(False)
        self.size_slider.setRange(4, 9)
        self.size_slider.setPageStep(1)
        self.size_slider.setValue(self.thumb_size / 20)
        self.size_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.size_slider.setMinimumWidth(140)
        self.size_slider.valueChanged.connect(self._new_thumb_size)
        layout.addWidget(self.size_slider, 1, 5)

    def set_drag_to_map(self, icon):
        self.drag_icon = icon

    def get_image(self, path):
        for image in self.images:
            if image.path == path:
                return image
        return None

    def get_images(self):
        return self.images

    def mousePressEvent(self, event):
        if self.scroll_area.underMouse():
            self._clear_selection()
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()

    @QtCore.pyqtSlot()
    def open_files(self):
        args = [
            self,
            self.tr('Open files'),
            self.app.config_store.get('paths', 'images', ''),
            self.tr("Images ({0});;Videos ({1});;All files (*)").format(
                ' '.join(['*.' + x for x in image_types()]),
                ' '.join(['*.' + x for x in video_types()]))
            ]
        if eval(self.app.config_store.get('pyqt', 'native_dialog', 'True')):
            pass
        elif qt_version_info >= (5, 0):
            args += [None, QtWidgets.QFileDialog.DontUseNativeDialog]
        else:
            args += [QtWidgets.QFileDialog.DontUseNativeDialog]
        path_list = QtWidgets.QFileDialog.getOpenFileNames(*args)
        if qt_version_info >= (5, 0):
            path_list = path_list[0]
        if not path_list:
            return
        # work around for Qt bug 33992
        # https://bugreports.qt-project.org/browse/QTBUG-33992
        if qt_version_info in ((4, 8, 4), (4, 8, 5)):
            path_list = list(map(unquote, path_list))
        self.open_file_list(path_list)

    @QtCore.pyqtSlot(list)
    def open_file_list(self, path_list):
        with Busy():
            for path in path_list:
                self.open_file(path)
        self.done_opening(path_list[-1])

    def open_file(self, path):
        path = os.path.abspath(path)
        if not os.path.isfile(path):
            return
        if self.get_image(path):
            # already opened this path
            return
        image = Image(path, self, thumb_size=self.thumb_size)
        self.images.append(image)
        self.show_thumbnail(image)

    def done_opening(self, path):
        self.app.config_store.set('paths', 'images', os.path.dirname(path))
        self._sort_thumbnails()

    def _date_key(self, image):
        result = image.metadata.date_taken
        if result is None:
            result = image.metadata.date_digitised
        if result is None:
            result = image.metadata.date_modified
        if result is None:
            # use file date as last resort
            result = datetime.fromtimestamp(os.path.getmtime(image.path))
        else:
            result = result.datetime
        # convert result to string and append path so photos with same
        # time stamp get sorted consistently
        result = result.isoformat() + image.path
        return result

    @QtCore.pyqtSlot()
    def _new_sort_order(self):
        self._sort_thumbnails()
        self.sort_order_changed.emit()

    def _sort_thumbnails(self):
        sort_date = self.sort_date.isChecked()
        self.app.config_store.set('controls', 'sort_date', str(sort_date))
        with Busy():
            if sort_date:
                self.images.sort(key=self._date_key)
            else:
                self.images.sort(key=lambda x: x.path)
            for image in self.images:
                self.show_thumbnail(image, False)
        if self.last_selected:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.last_selected)
        self.image_list_changed.emit()

    def show_thumbnail(self, image, live=True):
        self.thumbnails.layout().addWidget(image)
        if live:
            self.app.processEvents()
        image.load_thumbnail()
        if live:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(image)
            self.app.processEvents()

    def close_files(self, all_files):
        if not self.unsaved_files_dialog(all_files=all_files):
            return
        layout = self.thumbnails.layout()
        for image in list(self.images):
            if all_files or image.get_selected():
                self.images.remove(image)
                layout.removeWidget(image)
                image.setParent(None)
        self.last_selected = None
        self.selection_anchor = None
        self.emit_selection()
        self.image_list_changed.emit()

    @QtCore.pyqtSlot()
    def save_files(self, images=[]):
        if_mode = eval(self.app.config_store.get('files', 'image', 'True'))
        sc_mode = self.app.config_store.get('files', 'sidecar', 'auto')
        force_iptc = eval(
            self.app.config_store.get('files', 'force_iptc', 'False'))
        keep_time = eval(
            self.app.config_store.get('files', 'preserve_timestamps', 'False'))
        if not images:
            images = self.images
        with Busy():
            for image in images:
                if keep_time:
                    file_times = image.file_times
                else:
                    file_times = None
                image.metadata.save(if_mode, sc_mode, force_iptc, file_times)
        unsaved = False
        for image in self.images:
            if image.metadata.changed():
                unsaved = True
                break
        self.new_metadata.emit(unsaved)

    def unsaved_files_dialog(
            self, all_files=False, with_cancel=True, with_discard=True):
        """Return true if OK to continue with close or quit or whatever"""
        for image in self.images:
            if image.metadata.changed() and (all_files or image.selected):
                break
        else:
            return True
        dialog = QtWidgets.QMessageBox()
        dialog.setWindowTitle(self.tr('Photini: unsaved data'))
        dialog.setText(self.tr('<h3>Some images have unsaved metadata.</h3>'))
        dialog.setInformativeText(self.tr('Do you want to save your changes?'))
        dialog.setIcon(QtWidgets.QMessageBox.Warning)
        buttons = QtWidgets.QMessageBox.Save
        if with_cancel:
            buttons |= QtWidgets.QMessageBox.Cancel
        if with_discard:
            buttons |= QtWidgets.QMessageBox.Discard
        dialog.setStandardButtons(buttons)
        dialog.setDefaultButton(QtWidgets.QMessageBox.Save)
        result = dialog.exec_()
        if result == QtWidgets.QMessageBox.Save:
            self.save_files()
            return True
        return result == QtWidgets.QMessageBox.Discard

    def get_selected_images(self):
        selection = []
        for image in self.images:
            if image.get_selected():
                selection.append(image)
        return selection

    def emit_selection(self):
        self.selection_changed.emit(self.get_selected_images())

    def select_all(self):
        for image in self.images:
            image.set_selected(True)
        self.selection_anchor = None
        self.last_selected = None
        self.emit_selection()

    def move_to_prev_thumb(self):
        self._inc_selection(-1)

    def move_to_next_thumb(self):
        self._inc_selection(1)

    def move_to_first_thumb(self):
        self.select_image(self.images[0])

    def move_to_last_thumb(self):
        self.select_image(self.images[-1])

    def select_prev_thumb(self):
        self._inc_selection(-1, extend_selection=True)

    def select_next_thumb(self):
        self._inc_selection(1, extend_selection=True)

    def _inc_selection(self, inc, extend_selection=False):
        if self.last_selected:
            idx = self.images.index(self.last_selected)
            idx = (idx + inc) % len(self.images)
        else:
            idx = 0
        self.select_image(self.images[idx], extend_selection=extend_selection)

    @QtCore.pyqtSlot()
    def _new_thumb_size(self):
        self.thumb_size = self.size_slider.value() * 20
        self.app.config_store.set('controls', 'thumb_size', str(self.thumb_size))
        for image in self.images:
            image.set_thumb_size(self.thumb_size)
        if self.last_selected:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.last_selected)

    def select_image(
            self, image, extend_selection=False, multiple_selection=False):
        self.scroll_area.ensureWidgetVisible(image)
        if extend_selection and self.selection_anchor:
            idx1 = self.images.index(self.selection_anchor)
            idx2 = self.images.index(self.last_selected)
            for i in range(min(idx1, idx2), max(idx1, idx2) + 1):
                self.images[i].set_selected(False)
            idx2 = self.images.index(image)
            for i in range(min(idx1, idx2), max(idx1, idx2) + 1):
                self.images[i].set_selected(True)
        elif multiple_selection:
            image.set_selected(not image.get_selected())
            self.selection_anchor = image
        else:
            self._clear_selection()
            image.set_selected(True)
            self.selection_anchor = image
        self.last_selected = image
        self.emit_selection()

    def select_images(self, images):
        self._clear_selection()
        if not images:
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()
            return
        for image in images:
            image.set_selected(True)
            self.scroll_area.ensureWidgetVisible(image)
        self.selection_anchor = images[0]
        self.last_selected = images[-1]
        self.emit_selection()

    def _clear_selection(self):
        for image in self.images:
            if image.get_selected():
                image.set_selected(False)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2016-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import six
from collections import defaultdict
import logging
import math
import os
from six.moves.urllib.parse import unquote

import oauthlib
try:
    import PIL.Image as PIL
except ImportError:
    PIL = None
import pkg_resources
import requests
from requests_oauthlib import OAuth2Session
from requests_toolbelt import MultipartEncoder

from photini.configstore import key_store
from photini.pyqt import (
    Busy, MultiLineEdit, Qt, QtCore, QtGui, QtWebEngineWidgets,
    QtWebKitWidgets, QtWidgets, SingleLineEdit)
from photini.uploader import PhotiniUploader, UploaderSession

logger = logging.getLogger(__name__)
cities_cache = []

class FacebookSession(UploaderSession):
    name = 'facebook'
    scope = {
        'read' : ('user_photos',),
        'write': ('user_photos', 'publish_actions'),
        }

    def permitted(self, level):
        access_token = self.get_password()
        if not access_token:
            self.api = None
            return False
        if not self.api:
            self.api = OAuth2Session(token={'access_token': access_token})
        try:
            permissions = self.get('https://graph.facebook.com/me/permissions')
        except Exception:
            permissions = None
        if not permissions:
            return False
        for required in self.scope[level]:
            for permission in permissions['data']:
                if permission['permission'] != required:
                    continue
                if permission['status'] != 'granted':
                    return False
                break
            else:
                return False
        return True

    def get_auth_url(self, level):
        app_id = key_store.get('facebook', 'app_id')
        client = oauthlib.oauth2.MobileApplicationClient(app_id)
        self.api = OAuth2Session(
            client=client, scope=','.join(self.scope[level]),
            redirect_uri='https://www.facebook.com/connect/login_success.html',
            )
        result = self.api.authorization_url(
            'https://www.facebook.com/dialog/oauth',
            display='popup', auth_type='rerequest')[0]
        # use unquote to prevent "redirect_uri URL is not properly
        # formatted" error on Windows
        return unquote(result)

    def get_access_token(self, url, level):
        token = self.api.token_from_fragment(url)
        self.set_password(token['access_token'])
        self.api = None
        return self.permitted(level)

    def get_user(self):
        rsp = self.get('https://graph.facebook.com/me',
                       params={'fields': 'name,picture'})
        if not rsp:
            return None, None
        if rsp['picture']:
            url = rsp['picture']['data']['url']
            try:
                pic_rsp = self.api.get(url)
                pic_rsp.raise_for_status()
                return rsp['name'], pic_rsp.content
            except Exception as ex:
                logger.error('cannot read %s: %s', url, str(ex))
        return rsp['name'], None

    def get_album(self, album_id, fields):
        picture = None
        album = self.get(
            'https://graph.facebook.com/v2.6/' + album_id,
            params={'fields': fields})
        if not album:
            return {}, picture
        if 'cover_photo' in album:
            picture = self.get(
                'https://graph.facebook.com/v2.6/' + album['cover_photo']['id'],
                params={'fields': 'picture'})
        if picture:
            picture = picture['picture']
        return album, picture

    def get_albums(self, fields):
        albums = self.get(
            'https://graph.facebook.com/me/albums', params={'fields': fields})
        while True:
            if not albums:
                return
            for album in albums['data']:
                yield album
            if 'paging' not in albums or 'next' not in albums['paging']:
                return
            albums = self.get(albums['paging']['next'])

    def get_places(self, latlong, distance, query):
        places = self.get(
            'https://graph.facebook.com/search',
            params={'q'       : query,
                    'type'    : 'place',
                    'center'  : str(latlong),
                    'distance': distance,
                    'fields'  : 'category,id,location,name',
                    })
        while places:
            for place in places['data']:
                yield place
            if 'paging' not in places or 'next' not in places['paging']:
                return
            places = self.get(places['paging']['next'])

    def distance(self, a, b):
        # calculate approximate separation in metres of two nearby
        # lat/long values
        dx = (a.lon - b['longitude']) * 111320.0
        dy = (a.lat - b['latitude']) * 111320.0 * math.cos(math.radians(a.lat))
        return math.sqrt((dx * dx) + (dy * dy))

    def is_city(self, place):
        return (place['category'] in ('City', 'Neighborhood', 'Country',
                                      'State/province/region') and
                'latitude' in place['location'] and
                'longitude' in place['location'])

    def get_cities(self, latlong, city=None):
        # check cache for similar latlong
        nearest = None, 1.0e12
        for cache in cities_cache:
            dist = self.distance(latlong, cache['location'])
            if dist < nearest[1]:
                nearest = cache['cities_list'], dist
        if nearest[1] < 500:
            return nearest[0]
        elif nearest[1] < 1000:
            # initialise result to include nearby places
            result = list(nearest[0])
        else:
            result = []
        if city:
            # look for actual city
            for word in city.split(','):
                for place in self.get_places(latlong, 10000, word):
                    if place not in result and self.is_city(place):
                        result.append(place)
        else:
            # get a list of possible place names by searching for anything nearby
            hist = defaultdict(int)
            for place in self.get_places(latlong, 1000, ''):
                words = []
                if 'name' in place:
                    words += place['name'].split()
                if 'location' in place and 'city' in place['location']:
                    words += place['location']['city'].split()
                for word in words:
                    if word and word[0] in (',', '('):
                        word = word[1:]
                    if word and word[-1] in (',', ')'):
                        word = word[:-1]
                    if len(word) > 2:
                        hist[word.lower()] += 1
                if place not in result and self.is_city(place):
                    result.append(place)
            if not hist:
                hist['city'] = 1
            words = list(hist.keys())
            words.sort(key=lambda x: -hist[x])
            # search for cities, regions etc. using possible place names
            threshold = hist[words[len(words) // 2]]
            for word in words[:5]:
                if hist[word] <= threshold:
                    break
                for place in self.get_places(latlong, 10000, word):
                    if place not in result and self.is_city(place):
                        result.append(place)
        cities_cache.append({
            'location'   : {'latitude': latlong.lat, 'longitude': latlong.lon},
            'cities_list': result,
            })
        return result

    def do_upload(self, fileobj, image_type, image, params):
        fields = {
            'photo'   : ('source', fileobj),
            'no_story': str(params['no_story']),
            }
        title = image.metadata.title
        description = image.metadata.description
        if title and description:
            caption = title + '\n\n' + description
        elif title:
            caption = title
        elif description:
            caption = description
        else:
            caption = ''
        if caption:
            fields['caption'] = caption
        date_taken = image.metadata.date_taken
        if date_taken:
            fields['backdated_time'] = date_taken.datetime.isoformat()
            if date_taken.precision <= 5:
                fields['backdated_time_granularity'] = (
                    'year', 'month', 'day', 'hour', 'min')[date_taken.precision - 1]
        latlong = image.metadata.latlong
        if latlong and params['geo_tag']:
            city = None
            if image.metadata.location_taken:
                city = image.metadata.location_taken.city
            if image.metadata.location_shown:
                city = city or image.metadata.location_shown.city
            nearest = None, 1.0e12
            for place in self.get_cities(latlong, city=city):
                dist = self.distance(latlong, place['location'])
                if dist < nearest[1]:
                    nearest = place, dist
            nearest = nearest[0]
            if nearest:
                fields['place'] = nearest['id']
        data = MultipartEncoder(fields=fields)
        headers = {'Content-Type' : data.content_type}
        url = 'https://graph.facebook.com/v2.6/' + params['album_id'] + '/photos'
        try:
            self.post(url, data=data, headers=headers)
        except Exception as ex:
            return str(ex)
        return ''

    def get(self, *arg, **kw):
        rsp = self.api.get(*arg, **kw)
        rsp.raise_for_status()
        return rsp.json()

    def post(self, *arg, **kw):
        rsp = self.api.post(*arg, **kw)
        rsp.raise_for_status()
        return rsp.json()


if QtWebEngineWidgets:
    WebViewBase = QtWebEngineWidgets.QWebEngineView
else:
    WebViewBase = QtWebKitWidgets.QWebView

class WebView(WebViewBase):
    def sizeHint(self):
        return QtCore.QSize(580, 490)


class FacebookLoginPopup(QtWidgets.QDialog):
    def __init__(self, *arg, **kw):
        super(FacebookLoginPopup, self).__init__(*arg, **kw)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.browser = WebView()
        self.browser.urlChanged.connect(self.auth_url_changed)
        self.layout().addWidget(self.browser)
        buttons = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Cancel)
        buttons.rejected.connect(self.reject)
        self.layout().addWidget(buttons)

    def load_url(self, auth_url):
        self.browser.load(QtCore.QUrl(auth_url))

    @QtCore.pyqtSlot(QtCore.QUrl)
    def auth_url_changed(self, url):
        if url.path() != '/connect/login_success.html':
            return
        self.browser.setHtml('<p></p>')
        self.result = url.toString()
        if 'access_token' in url.fragment():
            return self.accept()
        self.reject()


class FacebookUploadConfig(QtWidgets.QWidget):
    new_album = QtCore.pyqtSignal()
    select_album = QtCore.pyqtSignal(int)

    def __init__(self, *arg, **kw):
        super(FacebookUploadConfig, self).__init__(*arg, **kw)
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.widgets = {}
        ## upload config
        config_group = QtWidgets.QGroupBox(self.tr('Options'))
        config_group.setLayout(QtWidgets.QFormLayout())
        self.layout().addWidget(config_group)
        # suppress feed story
        self.widgets['no_story'] = QtWidgets.QCheckBox()
        config_group.layout().addRow(
            self.tr('Suppress news feed story'), self.widgets['no_story'])
        label = config_group.layout().labelForField(self.widgets['no_story'])
        label.setWordWrap(True)
        label.setFixedWidth(90)
        # geotagging
        self.widgets['geo_tag'] = QtWidgets.QCheckBox()
        config_group.layout().addRow(
            self.tr('Set "city" from location metadata'), self.widgets['geo_tag'])
        self.widgets['geo_tag'].setChecked(True)
        label = config_group.layout().labelForField(self.widgets['geo_tag'])
        label.setWordWrap(True)
        label.setFixedWidth(90)
        # optimise
        self.widgets['optimise'] = QtWidgets.QCheckBox()
        config_group.layout().addRow(
            self.tr('Optimise image size'), self.widgets['optimise'])
        label = config_group.layout().labelForField(self.widgets['optimise'])
        label.setWordWrap(True)
        label.setFixedWidth(90)
        if PIL:
            self.widgets['optimise'].setChecked(True)
        else:
            self.widgets['optimise'].setEnabled(False)
            label.setEnabled(False)
        ## album details
        album_group = QtWidgets.QGroupBox(self.tr('Album'))
        album_group.setLayout(QtWidgets.QHBoxLayout())
        # left hand side
        album_form_left = QtWidgets.QFormLayout()
        album_form_left.setFieldGrowthPolicy(
            QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        album_group.layout().addLayout(album_form_left)
        # album title / selector
        self.widgets['album_choose'] = QtWidgets.QComboBox()
        self.widgets['album_choose'].activated.connect(self.select_album)
        album_form_left.addRow(self.tr('Title'), self.widgets['album_choose'])
        # album description
        self.widgets['album_description'] = QtWidgets.QPlainTextEdit()
        self.widgets['album_description'].setReadOnly(True)
        policy = self.widgets['album_description'].sizePolicy()
        policy.setVerticalStretch(1)
        self.widgets['album_description'].setSizePolicy(policy)
        album_form_left.addRow(self.tr('Description'), self.widgets['album_description'])
        # album location
        self.widgets['album_location'] = QtWidgets.QLineEdit()
        self.widgets['album_location'].setReadOnly(True)
        album_form_left.addRow(self.tr('Location'), self.widgets['album_location'])
        # right hand side
        album_form_right = QtWidgets.QVBoxLayout()
        album_group.layout().addLayout(album_form_right)
        # album thumbnail
        self.widgets['album_thumb'] = QtWidgets.QLabel()
        self.widgets['album_thumb'].setFixedSize(150, 150)
        self.widgets['album_thumb'].setAlignment(Qt.AlignHCenter | Qt.AlignTop)
        album_form_right.addWidget(self.widgets['album_thumb'])
        album_form_right.addStretch(1)
        # new album
        new_album_button = QtWidgets.QPushButton(self.tr('New album'))
        new_album_button.clicked.connect(self.new_album)
        album_form_right.addWidget(new_album_button)
        self.layout().addWidget(album_group, stretch=1)

    def show_album(self, album, picture):
        if 'description' in album:
            self.widgets['album_description'].setPlainText(album['description'])
        else:
            self.widgets['album_description'].clear()
        if 'location' in album:
            self.widgets['album_location'].setText(album['location'])
        else:
            self.widgets['album_location'].clear()
        pixmap = QtGui.QPixmap()
        if picture:
            rsp = requests.get(picture)
            if rsp.status_code == 200:
                pixmap.loadFromData(rsp.content)
            else:
                logger.error('HTTP error %d (%s)', rsp.status_code, picture)
        self.widgets['album_thumb'].setPixmap(pixmap)


class FacebookUploader(PhotiniUploader):
    session_factory = FacebookSession

    def __init__(self, *arg, **kw):
        self.upload_config = FacebookUploadConfig()
        super(FacebookUploader, self).__init__(self.upload_config, *arg, **kw)
        self.upload_config.new_album.connect(self.new_album)
        self.upload_config.select_album.connect(self.select_album)
        self.service_name = self.tr('Facebook')
        self.image_types = {
            'accepted': ('image/jpeg', 'image/png'),
            'rejected': ('image/x-portable-anymap', 'image/x-canon-cr2'),
            }
        self.login_popup = None
        # add Facebook icon to connect button
        icon_file = pkg_resources.resource_filename(
            'photini', 'data/facebook_logo.png')
        self.user_connect.setIcon(QtGui.QIcon(QtGui.QPixmap(icon_file)))

    def auth_dialog(self, auth_url):
        if not self.login_popup:
            # create dialog with embedded browser
            self.login_popup = FacebookLoginPopup(self)
            self.login_popup.setWindowTitle(
                self.tr('Photini: authorise {}').format(self.service_name))
        self.login_popup.load_url(auth_url)
        if self.login_popup.exec_() != QtWidgets.QDialog.Accepted:
            return None
        return self.login_popup.result

    def get_album_list(self, album_id=None):
        self.upload_config.widgets['album_choose'].clear()
        self.upload_config.widgets['album_choose'].addItem(
            self.tr('<default>'), 'me')
        if self.connected:
            selected = 0
            for album in self.session.get_albums('id,can_upload,name'):
                self.upload_config.widgets['album_choose'].addItem(
                    album['name'], album['id'])
                idx = self.upload_config.widgets['album_choose'].count() - 1
                if not album['can_upload']:
                    self.upload_config.widgets['album_choose'].setItemData(
                        idx, 0, Qt.UserRole - 1)
                elif album['id'] == album_id or not selected:
                    selected = idx
            self.upload_config.widgets['album_choose'].setCurrentIndex(selected)
            self.select_album(selected)
        else:
            self.upload_config.show_album({}, None)

    def optimise(self, image):
        try:
            im = PIL.open(image.path)
        except OSError:
            # PIL didn't recognise the file, so go via Qt
            path = self.convert_to_jpeg(image)
            im = PIL.open(path)
            im.load()
            os.unlink(path)
        # scale to one of Facebook's preferred sizes
        w, h = im.size
        old_size = max(w, h)
        new_size = None
        if old_size in (2048, 960, 720):
            pass
        elif old_size > 960:
            new_size = 2048
        elif old_size > 720:
            new_size = 960
        elif old_size > 360:
            new_size = 720
        if new_size:
            if w >= h:
                h = int((float(new_size * h) / float(w)) + 0.5)
                w = new_size
            else:
                w = int((float(new_size * w) / float(h)) + 0.5)
                h = new_size
            im = im.resize((w, h), PIL.ANTIALIAS)
        # save as temporary jpeg file
        path = self.get_temp_filename(image)
        im.save(path, format='jpeg', quality=95)
        # copy metadata although Facebook wipes most of it at present
        self.copy_metadata(image, path)
        return path

    def get_conversion_function(self, image):
        if (PIL and self.upload_config.widgets['optimise'].isChecked() and
                    self.is_convertible(image)):
            return self.optimise
        return super(FacebookUploader, self).get_conversion_function(image)

    def get_upload_params(self):
        idx = self.upload_config.widgets['album_choose'].currentIndex()
        return {
            'album_id': self.upload_config.widgets['album_choose'].itemData(idx),
            'no_story': self.upload_config.widgets['no_story'].isChecked(),
            'geo_tag' : self.upload_config.widgets['geo_tag'].isChecked(),
            }

    def upload_finished(self):
        # reload current album metadata (to update thumbnail)
        self.select_album(
            self.upload_config.widgets['album_choose'].currentIndex())

    @QtCore.pyqtSlot()
    def new_album(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(self.tr('Create new Facebook album'))
        dialog.setLayout(QtWidgets.QFormLayout())
        name = SingleLineEdit(spell_check=True)
        dialog.layout().addRow(self.tr('Title'), name)
        message = MultiLineEdit(spell_check=True)
        dialog.layout().addRow(self.tr('Description'), message)
        location = SingleLineEdit(spell_check=True)
        dialog.layout().addRow(self.tr('Location'), location)
        privacy = QtWidgets.QComboBox()
        for display_name, value in (
                (self.tr('Only me'),            '{value: "SELF"}'),
                (self.tr('All friends'),        '{value: "ALL_FRIENDS"}'),
                (self.tr('Friends of friends'), '{value: "FRIENDS_OF_FRIENDS"}'),
                (self.tr('Friends + networks'), '{value: "NETWORKS_FRIENDS"}'),
                (self.tr('Everyone'),           '{value: "EVERYONE"}'),
                ):
            privacy.addItem(display_name, value)
        dialog.layout().addRow(self.tr('Privacy'), privacy)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addRow(button_box)
        if dialog.exec_() != QtWidgets.QDialog.Accepted:
            return
        if not self.authorise('write'):
            self.refresh(force=True)
            return
        name = name.toPlainText().strip()
        if not name:
            return
        data = {'name': name}
        message = message.toPlainText().strip()
        if message:
            data['message'] = message
        location = location.toPlainText().strip()
        if location:
            data['location'] = location
        data['privacy'] = privacy.itemData(privacy.currentIndex())
        try:
            album = self.session.post(
                'https://graph.facebook.com/me/albums', data=data)
        except Exception as ex:
            logger.error(str(ex))
            self.refresh(force=True)
            return
        self.get_album_list(album_id=album['id'])

    @QtCore.pyqtSlot(int)
    def select_album(self, index):
        if not self.authorise('read'):
            self.refresh(force=True)
            return
        album_id = self.upload_config.widgets['album_choose'].itemData(index)
        if album_id == 'me':
            self.upload_config.show_album({}, None)
            return
        with Busy():
            album, picture = self.session.get_album(
                album_id, 'cover_photo,description,location,name')
            self.upload_config.show_album(album, picture)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import locale
import os
import webbrowser

import requests
import six

from photini.configstore import key_store
from photini.photinimap import PhotiniMap
from photini.pyqt import Busy, Qt, QtCore, QtWidgets, qt_version_info

class OpenStreetMap(PhotiniMap):
    def __init__(self, *arg, **kw):
        super(OpenStreetMap, self).__init__(*arg, **kw)
        self.block_timer = QtCore.QTimer(self)
        self.block_timer.setInterval(5000)
        self.block_timer.setSingleShot(True)
        self.block_timer.timeout.connect(self.enable_search)
        self.api_key = key_store.get('opencagedata', 'api_key')

    def get_page_elements(self):
        return {
            'head': '''
    <link rel="stylesheet"
      href="https://unpkg.com/leaflet@1.0.3/dist/leaflet.css" />
    <script type="text/javascript">
      var L_NO_TOUCH = true;
    </script>
    <script type="text/javascript"
      src="https://unpkg.com/leaflet@1.0.3/dist/leaflet.js">
    </script>
''',
            'body': '''
    <script type="text/javascript">
      initialize();
    </script>
''',
            }

    def show_terms(self):
        # return widget to display map terms and conditions
        layout = QtWidgets.QGridLayout()
        widget = QtWidgets.QPushButton(self.tr('Search powered by OpenCage'))
        widget.clicked.connect(self.load_tou_opencage)
        widget.setStyleSheet('QPushButton { font-size: 10px }')
        layout.addWidget(widget, 0, 0)
        widget = QtWidgets.QPushButton(self.tr('Map powered by Leaflet'))
        widget.clicked.connect(self.load_tou_leaflet)
        widget.setStyleSheet('QPushButton { font-size: 10px }')
        layout.addWidget(widget, 0, 1)
        if qt_version_info >= (5, 0):
            self.trUtf8 = self.tr
        widget = QtWidgets.QPushButton(
            self.trUtf8('Map data ©OpenStreetMap\ncontributors, licensed under ODbL'))
        widget.clicked.connect(self.load_tou_osm)
        widget.setStyleSheet('QPushButton { font-size: 10px }')
        layout.addWidget(widget, 1, 0)
        widget = QtWidgets.QPushButton(
            self.tr('Map tiles by CARTO\nlicensed under CC BY 3.0'))
        widget.clicked.connect(self.load_tou_tiles)
        widget.setStyleSheet('QPushButton { font-size: 10px }')
        layout.addWidget(widget, 1, 1)
        return layout

    @QtCore.pyqtSlot()
    def load_tou_opencage(self):
        webbrowser.open_new('https://geocoder.opencagedata.com/')

    @QtCore.pyqtSlot()
    def load_tou_leaflet(self):
        webbrowser.open_new('http://leafletjs.com/')

    @QtCore.pyqtSlot()
    def load_tou_osm(self):
        webbrowser.open_new('http://www.openstreetmap.org/copyright')

    @QtCore.pyqtSlot()
    def load_tou_tiles(self):
        webbrowser.open_new('https://carto.com/attribution')

    @QtCore.pyqtSlot()
    def enable_search(self):
        self.edit_box.lineEdit().setEnabled(self.map_loaded)
        if self.search_string:
            item = self.edit_box.model().item(1)
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        self.display_coords()

    def disable_search(self):
        self.edit_box.lineEdit().setEnabled(False)
        if self.search_string:
            item = self.edit_box.model().item(1)
            item.setFlags(~(Qt.ItemIsSelectable | Qt.ItemIsEnabled))
        self.auto_location.setEnabled(False)
        self.block_timer.start()

    def do_search(self, query, params={}):
        self.disable_search()
        params['key'] = self.api_key
        params['q'] = query
        params['abbrv'] = '1'
        params['no_annotations'] = '1'
        lang, encoding = locale.getdefaultlocale()
        if lang:
            params['language'] = lang
        with Busy():
            try:
                rsp = requests.get(
                    'https://api.opencagedata.com/geocode/v1/json',
                    params=params)
            except Exception as ex:
                self.logger.error(str(ex))
                return None
        if rsp.status_code >= 400:
            self.logger.error('Search error %d', rsp.status_code)
            return None
        rsp = rsp.json()
        status = rsp['status']
        if status['code'] != 200:
            self.logger.error(
                'Search error %d: %s', status['code'], status['message'])
            return None
        rate = rsp['rate']
        self.block_timer.setInterval(
            5000 * rate['limit'] // max(rate['remaining'], 1))
        return rsp

    @QtCore.pyqtSlot()
    def get_address(self):
        lat, lon = self.coords.get_value().split(',')
        rsp = self.do_search(lat.strip() + ' ' + lon.strip())
        if not rsp:
            return
        if rsp['total_results'] < 1:
            self.logger.error('Address not found')
            return
        address = rsp['results'][0]['components']
        location = []
        for iptc_key, osm_keys in (
                ('world_region',   ('continent',)),
                ('country_code',   ('country_code', 'ISO_3166-1_alpha-2')),
                ('country_name',   ('country',)),
                ('province_state', ('region', 'county',
                                    'state_district', 'state')),
                ('city',           ('hamlet', 'locality', 'neighbourhood',
                                    'village', 'suburb',
                                    'town', 'city_district', 'city')),
                ('sublocation',    ('building', 'house_number',
                                    'footway', 'pedestrian', 'road'))):
            element = []
            for key in osm_keys:
                if key not in address:
                    continue
                if iptc_key == 'country_code':
                    element = [address[key]]
                elif address[key] not in element:
                    element.append(address[key])
                del(address[key])
            location.append(', '.join(element))
        # put remaining keys in sublocation
        for key in address:
            if key in ('postcode', 'state_code', '_type'):
                continue
            location[-1] = '{}: {}, {}'.format(key, address[key], location[-1])
        self.set_location_taken(*location)

    @QtCore.pyqtSlot()
    def search(self, search_string=None):
        if not search_string:
            search_string = self.edit_box.lineEdit().text()
            self.edit_box.clearEditText()
        if not search_string:
            return
        self.search_string = search_string
        self.clear_search()
        bounds = self.map_status['bounds']
        scale = 2 ** (self.map_status['zoom'] - 9)
        if scale >= 1:
            w = (bounds[1] - bounds[3]) * scale
            h = (bounds[0] - bounds[2]) * scale
            lat, lon = self.map_status['centre']
            bounds = (lon - w, max(lat - h, -90.0),
                      lon + w, min(lat + h,  90.0))
        else:
            bounds = (bounds[3], bounds[2], bounds[1], bounds[0])
        rsp = self.do_search(
            search_string, {'bounds': ','.join(map(str, bounds))})
        if not rsp:
            return
        if not rsp['results']:
            # repeat search without bounds
            rsp = self.do_search(search_string)
            if not rsp:
                return
        for result in rsp['results']:
            self.search_result(
                result['bounds']['northeast']['lat'],
                result['bounds']['northeast']['lng'],
                result['bounds']['southwest']['lat'],
                result['bounds']['southwest']['lng'],
                result['formatted'])

    @QtCore.pyqtSlot(int)
    def marker_drag_start(self, marker_id):
        blocked = self.image_list.blockSignals(True)
        self.image_list.select_images(self.marker_images[marker_id])
        self.image_list.blockSignals(blocked)
        self.coords.setEnabled(True)
        for other_id, images in self.marker_images.items():
            if other_id != marker_id:
                self.JavaScript('enableMarker({:d},{:d})'.format(
                    other_id, False))
        self.display_coords()



# -*- coding: utf-8 -*-
#  Photini - a simple photo metadata editor.
#  http://github.com/jim-easterbrook/Photini
#  Copyright (C) 2012-17  Jim Easterbrook  jim@jim-easterbrook.me.uk
#
#  This program is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

try:
    import keyring
except (ImportError, RuntimeError):
    keyring = None

from photini.pyqt import Qt, QtCore, QtWidgets

class EditSettings(QtWidgets.QDialog):
    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent)
        self.config_store = QtWidgets.QApplication.instance().config_store
        self.setWindowTitle(self.tr('Photini: settings'))
        self.setLayout(QtWidgets.QVBoxLayout())
        # main dialog area
        scroll_area = QtWidgets.QScrollArea()
        self.layout().addWidget(scroll_area)
        panel = QtWidgets.QWidget()
        panel.setLayout(QtWidgets.QFormLayout())
        panel.layout().setRowWrapPolicy(max(QtWidgets.QFormLayout.WrapLongRows,
                                            panel.layout().rowWrapPolicy()))
        # apply & cancel buttons
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Apply | QtWidgets.QDialogButtonBox.Cancel)
        self.button_box.clicked.connect(self.button_clicked)
        self.layout().addWidget(self.button_box)
        # copyright holder name
        self.copyright_name = QtWidgets.QLineEdit()
        self.copyright_name.setText(
            self.config_store.get('user', 'copyright_name', ''))
        self.copyright_name.setMinimumWidth(200)
        panel.layout().addRow(self.tr('Copyright holder name'), self.copyright_name)
        # copyright text
        self.copyright_text = QtWidgets.QLineEdit()
        self.copyright_text.setText(
            self.config_store.get('user', 'copyright_text', ''))
        self.copyright_name.setMinimumWidth(300)
        panel.layout().addRow(self.tr('Copyright text'), self.copyright_text)
        # creator name
        self.creator_name = QtWidgets.QLineEdit()
        self.creator_name.setText(
            self.config_store.get('user', 'creator_name', ''))
        panel.layout().addRow(self.tr('Creator name'), self.creator_name)
        # reset flickr
        self.reset_flickr = QtWidgets.QCheckBox()
        panel.layout().addRow(self.tr('Disconnect from Flickr'), self.reset_flickr)
        if not keyring or keyring.get_password('photini', 'flickr') is None:
            self.reset_flickr.setDisabled(True)
            panel.layout().labelForField(self.reset_flickr).setDisabled(True)
        # reset picasa
        self.reset_picasa = QtWidgets.QCheckBox()
        panel.layout().addRow(self.tr('Disconnect from Google Photos'), self.reset_picasa)
        if not keyring or keyring.get_password('photini', 'picasa') is None:
            self.reset_picasa.setDisabled(True)
            panel.layout().labelForField(self.reset_picasa).setDisabled(True)
        # reset facebook
        self.reset_facebook = QtWidgets.QCheckBox()
        panel.layout().addRow(self.tr('Disconnect from Facebook'), self.reset_facebook)
        if not keyring or keyring.get_password('photini', 'facebook') is None:
            self.reset_facebook.setDisabled(True)
            panel.layout().labelForField(self.reset_facebook).setDisabled(True)
        # IPTC data
        force_iptc = eval(self.config_store.get('files', 'force_iptc', 'False'))
        self.write_iptc = QtWidgets.QCheckBox(self.tr('Write unconditionally'))
        self.write_iptc.setChecked(force_iptc)
        panel.layout().addRow(self.tr('IPTC metadata'), self.write_iptc)
        # sidecar files
        if_mode = eval(self.config_store.get('files', 'image', 'True'))
        sc_mode = self.config_store.get('files', 'sidecar', 'auto')
        if not if_mode:
            sc_mode = 'always'
        self.sc_always = QtWidgets.QRadioButton(self.tr('Always create'))
        self.sc_always.setChecked(sc_mode == 'always')
        panel.layout().addRow(self.tr('Sidecar files'), self.sc_always)
        self.sc_auto = QtWidgets.QRadioButton(self.tr('Create if necessary'))
        self.sc_auto.setChecked(sc_mode == 'auto')
        self.sc_auto.setEnabled(if_mode)
        panel.layout().addRow('', self.sc_auto)
        self.sc_delete = QtWidgets.QRadioButton(self.tr('Delete when possible'))
        self.sc_delete.setChecked(sc_mode == 'delete')
        self.sc_delete.setEnabled(if_mode)
        panel.layout().addRow('', self.sc_delete)
        # image file locking
        self.write_if = QtWidgets.QCheckBox(self.tr('(when possible)'))
        self.write_if.setChecked(if_mode)
        self.write_if.clicked.connect(self.new_write_if)
        panel.layout().addRow(self.tr('Write to image'), self.write_if)
        # preserve file timestamps
        keep_time = eval(
            self.config_store.get('files', 'preserve_timestamps', 'False'))
        self.keep_time = QtWidgets.QCheckBox()
        self.keep_time.setChecked(keep_time)
        panel.layout().addRow(
            self.tr('Preserve file timestamps'), self.keep_time)
        # add panel to scroll area after its size is known
        scroll_area.setWidget(panel)

    @QtCore.pyqtSlot()
    def new_write_if(self):
        if_mode = self.write_if.isChecked()
        self.sc_auto.setEnabled(if_mode)
        self.sc_delete.setEnabled(if_mode)
        if not if_mode:
            self.sc_always.setChecked(True)

    @QtCore.pyqtSlot(QtWidgets.QAbstractButton)
    def button_clicked(self, button):
        if button != self.button_box.button(QtWidgets.QDialogButtonBox.Apply):
            return self.reject()
        # change config
        self.config_store.set('user', 'copyright_name', self.copyright_name.text())
        self.config_store.set('user', 'copyright_text', self.copyright_text.text())
        self.config_store.set('user', 'creator_name', self.creator_name.text())
        if (self.reset_flickr.isChecked() and
                            keyring.get_password('photini', 'flickr')):
            keyring.delete_password('photini', 'flickr')
        if (self.reset_picasa.isChecked() and
                            keyring.get_password('photini', 'picasa')):
            keyring.delete_password('photini', 'picasa')
        if (self.reset_facebook.isChecked() and
                            keyring.get_password('photini', 'facebook')):
            keyring.delete_password('photini', 'facebook')
        self.config_store.set(
            'files', 'force_iptc', str(self.write_iptc.isChecked()))
        if self.sc_always.isChecked():
            sc_mode = 'always'
        elif self.sc_auto.isChecked():
            sc_mode = 'auto'
        else:
            sc_mode = 'delete'
        self.config_store.set('files', 'sidecar', sc_mode)
        self.config_store.set('files', 'image', str(self.write_if.isChecked()))
        self.config_store.set(
            'files', 'preserve_timestamps', str(self.keep_time.isChecked()))
        return self.accept()
